# --------------------------------------------------------------------

cdef class Section(Object):
    """Mapping from integers in a range to unstructured set of integers."""

    def __cinit__(self):
        self.obj = <PetscObject*> &self.sec
        self.sec  = NULL

    def view(self, Viewer viewer=None) -> None:
        """View the section.

        Collective.

        Parameters
        ----------
        viewer
            A `Viewer` to display the section.

        See Also
        --------
        petsc.PetscSectionView

        """
        cdef PetscViewer vwr = NULL
        if viewer is not None: vwr = viewer.vwr
        CHKERR(PetscSectionView(self.sec, vwr))

    def destroy(self) -> Self:
        """Destroy a section.

        Not collective.

        See Also
        --------
        petsc.PetscSectionDestroy

        """
        CHKERR(PetscSectionDestroy(&self.sec))
        return self

    def create(self, comm: Comm | None = None) -> Self:
        """Allocate a section and set the map contents to the default.

        Collective.

        Typical calling sequence:
        - `create`
        - `setNumFields`
        - `setChart`
        - `setDof`
        - `setUp`
        - `getOffset`
        - `destroy`

        The `Section` object and methods are intended to be used in the PETSc
        Vec and Mat implementations. The indices returned by the `Section` are
        appropriate for the kind of `Vec` it is associated with. For example,
        if the vector being indexed is a local vector, we call the section a
        local section. If the section indexes a global vector, we call it a
        global section. For parallel vectors, like global vectors, we use
        negative indices to indicate DOFs owned by other processes.

        Parameters
        ----------
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        petsc.PetscSectionCreate

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef PetscSection newsec = NULL
        CHKERR(PetscSectionCreate(ccomm, &newsec))
        CHKERR(PetscCLEAR(self.obj)); self.sec = newsec
        return self

    def clone(self) -> Section:
        """Return a copy of the section.

        Collective.

        The copy is shallow, if possible.

        See Also
        --------
        petsc.PetscSectionClone

        """
        cdef Section sec = <Section>type(self)()
        CHKERR(PetscSectionClone(self.sec, &sec.sec))
        return sec

    def setUp(self) -> None:
        """Calculate offsets.

        Not collective.

        Offsets are based on the number of degrees of freedom for each point.

        See Also
        --------
        petsc.PetscSectionSetUp

        """
        CHKERR(PetscSectionSetUp(self.sec))

    def reset(self) -> None:
        """Free all section data.

        Not collective.

        See Also
        --------
        petsc.PetscSectionReset

        """
        CHKERR(PetscSectionReset(self.sec))

    def getNumFields(self) -> int:
        """Return the number of fields in a section.

        Not collective.

        Returns ``0`` if no fields were defined.

        See Also
        --------
        setNumFields, petsc.PetscSectionGetNumFields

        """
        cdef PetscInt numFields = 0
        CHKERR(PetscSectionGetNumFields(self.sec, &numFields))
        return toInt(numFields)

    def setNumFields(self, numFields: int) -> None:
        """Set the number of fields in a section.

        Not collective.

        Parameters
        ----------
        numFields
            The number of fields.

        See Also
        --------
        getNumFields, petsc.PetscSectionSetNumFields

        """
        cdef PetscInt cnumFields = asInt(numFields)
        CHKERR(PetscSectionSetNumFields(self.sec, cnumFields))

    def getFieldName(self, field: int) -> str:
        """Return the name of a field in the section.

        Not collective.

        Parameters
        ----------
        field
            The field number.

        See Also
        --------
        setFieldName, petsc.PetscSectionGetFieldName

        """
        cdef PetscInt cfield = asInt(field)
        cdef const char *fieldName = NULL
        CHKERR(PetscSectionGetFieldName(self.sec, cfield, &fieldName))
        return bytes2str(fieldName)

    def setFieldName(self, field: int, fieldName: str) -> None:
        """Set the name of a field in the section.

        Not collective.

        Parameters
        ----------
        field
            The field number.
        fieldName
            The field name.

        See Also
        --------
        getFieldName, petsc.PetscSectionSetFieldName

        """
        cdef PetscInt cfield = asInt(field)
        cdef const char *cname = NULL
        fieldName = str2bytes(fieldName, &cname)
        CHKERR(PetscSectionSetFieldName(self.sec, cfield, cname))

    def getFieldComponents(self, field: int) -> int:
        """Return the number of field components for the given field.

        Not collective.

        Parameters
        ----------
        field
            The field number.

        See Also
        --------
        setFieldComponents, petsc.PetscSectionGetFieldComponents

        """
        cdef PetscInt cfield = asInt(field), cnumComp = 0
        CHKERR(PetscSectionGetFieldComponents(self.sec, cfield, &cnumComp))
        return toInt(cnumComp)

    def setFieldComponents(self, field: int, numComp: int) -> None:
        """Set the number of field components for the given field.

        Not collective.

        Parameters
        ----------
        field
            The field number.
        numComp
            The number of field components.

        See Also
        --------
        getFieldComponents, petsc.PetscSectionSetFieldComponents

        """
        cdef PetscInt cfield = asInt(field)
        cdef PetscInt cnumComp = asInt(numComp)
        CHKERR(PetscSectionSetFieldComponents(self.sec, cfield, cnumComp))

    def getChart(self) -> tuple[int, int]:
        """Return the range in which points (indices) lie for this section.

        Not collective.

        The range is [pStart, pEnd), i.e., from the first point to one past the
        last point.

        See Also
        --------
        petsc.PetscSectionGetChart

        """
        cdef PetscInt pStart = 0, pEnd = 0
        CHKERR(PetscSectionGetChart(self.sec, &pStart, &pEnd))
        return toInt(pStart), toInt(pEnd)

    def setChart(self, pStart: int, pEnd: int) -> None:
        """Set the range in which points (indices) lie for this section.

        Not collective.

        The range is [pStart, pEnd), i.e., from the first point to one past the
        last point.

        Parameters
        ----------
        pStart
            The first point.
        pEnd
            One past the last point.

        See Also
        --------
        petsc.PetscSectionSetChart

        """
        cdef PetscInt cStart = asInt(pStart)
        cdef PetscInt cEnd   = asInt(pEnd)
        CHKERR(PetscSectionSetChart(self.sec, cStart, cEnd))

    def getPermutation(self) -> IS:
        """Return the permutation that was set with `setPermutation`.

        Not collective.

        See Also
        --------
        setPermutation, petsc.PetscSectionGetPermutation

        """
        cdef IS perm = IS()
        CHKERR(PetscSectionGetPermutation(self.sec, &perm.iset))
        CHKERR(PetscINCREF(perm.obj))
        return perm

    def setPermutation(self, IS perm) -> None:
        """Set the permutation for [0, pEnd - pStart).

        Not collective.

        Parameters
        ----------
        perm
            The permutation of points.

        See Also
        --------
        getPermutation, petsc.PetscSectionSetPermutation

        """
        CHKERR(PetscSectionSetPermutation(self.sec, perm.iset))

    def getDof(self, point: int) -> int:
        """Return the number of degrees of freedom for a given point.

        Not collective.

        In a global section, this value will be negative for points not owned
        by this process.

        Parameters
        ----------
        point
            The point.

        See Also
        --------
        setDof, addDof, petsc.PetscSectionGetDof

        """
        cdef PetscInt cpoint = asInt(point), cnumDof = 0
        CHKERR(PetscSectionGetDof(self.sec, cpoint, &cnumDof))
        return toInt(cnumDof)

    def setDof(self, point: int, numDof: int) -> None:
        """Set the number of degrees of freedom associated with a given point.

        Not collective.

        Parameters
        ----------
        point
            The point.
        numDof
            The number of DOFs.

        See Also
        --------
        getDof, addDof, petsc.PetscSectionSetDof

        """
        cdef PetscInt cpoint = asInt(point)
        cdef PetscInt cnumDof = asInt(numDof)
        CHKERR(PetscSectionSetDof(self.sec, cpoint, cnumDof))

    def addDof(self, point: int, numDof: int) -> None:
        """Add ``numDof`` degrees of freedom associated with a given point.

        Not collective.

        Parameters
        ----------
        point
            The point.
        numDof
            The number of additional DOFs.

        See Also
        --------
        setDof, getDof, petsc.PetscSectionAddDof

        """
        cdef PetscInt cpoint = asInt(point)
        cdef PetscInt cnumDof = asInt(numDof)
        CHKERR(PetscSectionAddDof(self.sec, cpoint, cnumDof))

    def getFieldDof(self, point: int, field: int) -> int:
        """Return the number of DOFs associated with a field on a given point.

        Not collective.

        Parameters
        ----------
        point
            The point.
        field
            The field.

        See Also
        --------
        setFieldDof, petsc.PetscSectionGetFieldDof

        """
        cdef PetscInt cpoint = asInt(point), cnumDof = 0
        cdef PetscInt cfield = asInt(field)
        CHKERR(PetscSectionGetFieldDof(self.sec, cpoint, cfield, &cnumDof))
        return toInt(cnumDof)

    def setFieldDof(self, point: int, field: int, numDof: int) -> None:
        """Set the number of DOFs associated with a field on a given point.

        Not collective.

        Parameters
        ----------
        point
            The point.
        field
            The field.
        numDof
            The number of DOFs.

        See Also
        --------
        getFieldDof, addFieldDof, petsc.PetscSectionSetFieldDof

        """
        cdef PetscInt cpoint = asInt(point)
        cdef PetscInt cfield = asInt(field)
        cdef PetscInt cnumDof = asInt(numDof)
        CHKERR(PetscSectionSetFieldDof(self.sec, cpoint, cfield, cnumDof))

    def addFieldDof(self, point: int, field: int, numDof: int) -> None:
        """Add ``numDof`` DOFs associated with a field on a given point.

        Not collective.

        Parameters
        ----------
        point
            The point.
        field
            The field.
        numDof
            The number of additional DOFs.

        See Also
        --------
        setFieldDof, getFieldDof, petsc.PetscSectionAddFieldDof

        """
        cdef PetscInt cpoint = asInt(point)
        cdef PetscInt cfield = asInt(field)
        cdef PetscInt cnumDof = asInt(numDof)
        CHKERR(PetscSectionAddFieldDof(self.sec, cpoint, cfield, cnumDof))

    def getConstraintDof(self, point: int) -> int:
        """Return the number of constrained DOFs associated with a given point.

        Not collective.

        Parameters
        ----------
        point
            The point.

        See Also
        --------
        setConstraintDof, petsc.PetscSectionGetConstraintDof

        """
        cdef PetscInt cpoint = asInt(point), cnumDof = 0
        CHKERR(PetscSectionGetConstraintDof(self.sec, cpoint, &cnumDof))
        return toInt(cnumDof)

    def setConstraintDof(self, point: int, numDof: int) -> None:
        """Set the number of constrained DOFs associated with a given point.

        Not collective.

        Parameters
        ----------
        point
            The point.
        numDof
            The number of DOFs which are fixed by constraints.

        See Also
        --------
        getConstraintDof, addConstraintDof, petsc.PetscSectionSetConstraintDof

        """
        cdef PetscInt cpoint = asInt(point)
        cdef PetscInt cnumDof = asInt(numDof)
        CHKERR(PetscSectionSetConstraintDof(self.sec, cpoint, cnumDof))

    def addConstraintDof(self, point: int, numDof: int) -> None:
        """Increment the number of constrained DOFs for a given point.

        Not collective.

        Parameters
        ----------
        point
            The point.
        numDof
            The number of additional DOFs which are fixed by constraints.

        See Also
        --------
        setConstraintDof, getConstraintDof, petsc.PetscSectionAddConstraintDof

        """
        cdef PetscInt cpoint = asInt(point)
        cdef PetscInt cnumDof = asInt(numDof)
        CHKERR(PetscSectionAddConstraintDof(self.sec, cpoint, cnumDof))

    def getFieldConstraintDof(self, point: int, field: int) -> int:
        """Return the number of constrained DOFs for a given field on a point.

        Not collective.

        Parameters
        ----------
        point
            The point.
        field
            The field.

        See Also
        --------
        setFieldConstraintDof, petsc.PetscSectionGetFieldConstraintDof

        """
        cdef PetscInt cpoint = asInt(point), cnumDof = 0
        cdef PetscInt cfield = asInt(field)
        CHKERR(PetscSectionGetFieldConstraintDof(self.sec, cpoint, cfield, &cnumDof))
        return toInt(cnumDof)

    def setFieldConstraintDof(
        self,
        point: int,
        field: int,
        numDof: int) -> None:
        """Set the number of constrained DOFs for a given field on a point.

        Not collective.

        Parameters
        ----------
        point
            The point.
        field
            The field.
        numDof
            The number of DOFs which are fixed by constraints.

        See Also
        --------
        getFieldConstraintDof, addFieldConstraintDof
        petsc.PetscSectionSetFieldConstraintDof

        """
        cdef PetscInt cpoint = asInt(point)
        cdef PetscInt cfield = asInt(field)
        cdef PetscInt cnumDof = asInt(numDof)
        CHKERR(PetscSectionSetFieldConstraintDof(self.sec, cpoint, cfield, cnumDof))

    def addFieldConstraintDof(
        self,
        point: int,
        field: int,
        numDof: int) -> None:
        """Add ``numDof`` constrained DOFs for a given field on a point.

        Not collective.

        Parameters
        ----------
        point
            The point.
        field
            The field.
        numDof
            The number of additional DOFs which are fixed by constraints.

        See Also
        --------
        setFieldConstraintDof, getFieldConstraintDof
        petsc.PetscSectionAddFieldConstraintDof

        """
        cdef PetscInt cpoint = asInt(point)
        cdef PetscInt cfield = asInt(field)
        cdef PetscInt cnumDof = asInt(numDof)
        CHKERR(PetscSectionAddFieldConstraintDof(self.sec, cpoint, cfield, cnumDof))

    def getConstraintIndices(self, point: int) -> ArrayInt:
        """Return the point DOFs numbers which are constrained for a given point.

        Not collective.

        The range is in [0, DOFs).

        Parameters
        ----------
        point
            The point.

        See Also
        --------
        setConstraintIndices, petsc.PetscSectionGetConstraintIndices

        """
        cdef PetscInt cpoint = asInt(point)
        cdef PetscInt nindex = 0
        cdef const PetscInt *indices = NULL
        CHKERR(PetscSectionGetConstraintDof(self.sec, cpoint, &nindex))
        CHKERR(PetscSectionGetConstraintIndices(self.sec, cpoint, &indices))
        return array_i(nindex, indices)

    def setConstraintIndices(self, point: int, indices: Sequence[int]) -> None:
        """Set the point DOFs numbers, in [0, DOFs), which are constrained.

        Not collective.

        Parameters
        ----------
        point
            The point.
        indices
            The constrained DOFs.

        See Also
        --------
        getConstraintIndices, petsc.PetscSectionSetConstraintIndices

        """
        cdef PetscInt cpoint = asInt(point)
        cdef PetscInt nindex = 0
        cdef PetscInt *cindices = NULL
        indices = iarray_i(indices, &nindex, &cindices)
        CHKERR(PetscSectionSetConstraintDof(self.sec, cpoint, nindex))
        CHKERR(PetscSectionSetConstraintIndices(self.sec, cpoint, cindices))

    def getFieldConstraintIndices(self, point: int, field: int) -> ArrayInt:
        """Return the field DOFs numbers, in [0, DOFs), which are constrained.

        Not collective.

        The constrained DOFs are sorted in ascending order.

        Parameters
        ----------
        field
            The field number.
        point
            The point.

        See Also
        --------
        setFieldConstraintIndices, petsc.PetscSectionGetFieldConstraintIndices

        """
        cdef PetscInt cpoint = asInt(point)
        cdef PetscInt cfield = asInt(field)
        cdef PetscInt nindex = 0
        cdef const PetscInt *indices = NULL
        CHKERR(PetscSectionGetFieldConstraintDof(self.sec, cpoint, cfield, &nindex))
        CHKERR(PetscSectionGetFieldConstraintIndices(self.sec, cpoint, cfield, &indices))
        return array_i(nindex, indices)

    def setFieldConstraintIndices(
        self,
        point: int,
        field: int,
        indices: Sequence[int]) -> None:
        """Set the field DOFs numbers, in [0, DOFs), which are constrained.

        Not collective.

        Parameters
        ----------
        point
            The point.
        field
            The field number.
        indices
            The constrained DOFs.

        See Also
        --------
        getFieldConstraintIndices, petsc.PetscSectionSetFieldConstraintIndices

        """
        cdef PetscInt cpoint = asInt(point)
        cdef PetscInt cfield = asInt(field)
        cdef PetscInt nindex = 0
        cdef PetscInt *cindices = NULL
        indices = iarray_i(indices, &nindex, &cindices)
        CHKERR(PetscSectionSetFieldConstraintDof(self.sec, cpoint, cfield, nindex))
        CHKERR(PetscSectionSetFieldConstraintIndices(self.sec, cpoint, cfield, cindices))

    def getMaxDof(self) -> int:
        """Return the maximum number of DOFs for any point in the section.

        Not collective.

        See Also
        --------
        petsc.PetscSectionGetMaxDof

        """
        cdef PetscInt maxDof = 0
        CHKERR(PetscSectionGetMaxDof(self.sec, &maxDof))
        return toInt(maxDof)

    def getStorageSize(self) -> int:
        """Return the size capable of holding all the DOFs defined in a section.

        Not collective.

        See Also
        --------
        getConstrainedStorageSize, petsc.PetscSectionGetStorageSize

        """
        cdef PetscInt size = 0
        CHKERR(PetscSectionGetStorageSize(self.sec, &size))
        return toInt(size)

    def getConstrainedStorageSize(self) -> int:
        """Return the size capable of holding all unconstrained DOFs in a section.

        Not collective.

        See Also
        --------
        getStorageSize, petsc.PetscSectionGetConstrainedStorageSize

        """
        cdef PetscInt size = 0
        CHKERR(PetscSectionGetConstrainedStorageSize(self.sec, &size))
        return toInt(size)

    def getOffset(self, point: int) -> int:
        """Return the offset for the DOFs associated with the given point.

        Not collective.

        In a global section, this offset will be negative for points not owned
        by this process.

        Parameters
        ----------
        point
            The point.

        See Also
        --------
        setOffset, petsc.PetscSectionGetOffset

        """
        cdef PetscInt cpoint = asInt(point), offset = 0
        CHKERR(PetscSectionGetOffset(self.sec, cpoint, &offset))
        return toInt(offset)

    def setOffset(self, point: int, offset: int) -> None:
        """Set the offset for the DOFs associated with the given point.

        Not collective.

        The user usually does not call this function, but uses `setUp`.

        Parameters
        ----------
        point
            The point.
        offset
            The offset.

        See Also
        --------
        getOffset, petsc.PetscSectionSetOffset

        """
        cdef PetscInt cpoint = asInt(point)
        cdef PetscInt coffset = asInt(offset)
        CHKERR(PetscSectionSetOffset(self.sec, cpoint, coffset))

    def getFieldOffset(self, point: int, field: int) -> int:
        """Return the offset for the field DOFs on the given point.

        Not collective.

        In a global section, this offset will be negative for points not owned
        by this process.

        Parameters
        ----------
        point
            The point.
        field
            The field.

        See Also
        --------
        setFieldOffset, petsc.PetscSectionGetFieldOffset

        """
        cdef PetscInt cpoint = asInt(point)
        cdef PetscInt cfield = asInt(field)
        cdef PetscInt offset = 0
        CHKERR(PetscSectionGetFieldOffset(self.sec, cpoint, cfield, &offset))
        return toInt(offset)

    def setFieldOffset(self, point: int, field: int, offset: int) -> None:
        """Set the offset for the DOFs on the given field at a point.

        Not collective.

        The user usually does not call this function, but uses `setUp`.

        Parameters
        ----------
        point
            The point.
        field
            The field.
        offset
            The offset.

        See Also
        --------
        getFieldOffset, petsc.PetscSectionSetFieldOffset

        """
        cdef PetscInt cpoint = asInt(point)
        cdef PetscInt cfield = asInt(field)
        cdef PetscInt coffset = asInt(offset)
        CHKERR(PetscSectionSetFieldOffset(self.sec, cpoint, cfield, coffset))

    def getOffsetRange(self) -> tuple[int, int]:
        """Return the full range of offsets, [start, end), for a section.

        Not collective.

        See Also
        --------
        petsc.PetscSectionGetOffsetRange

        """
        cdef PetscInt oStart = 0, oEnd = 0
        CHKERR(PetscSectionGetOffsetRange(self.sec, &oStart, &oEnd))
        return toInt(oStart), toInt(oEnd)

    # FIXME: Hardcoded PETSC_FALSE parameters
    def createGlobalSection(self, SF sf) -> Section:
        """Create a section describing the global field layout.

        Collective.

        The section describes the global field layout using the local section
        and an `SF` describing the section point overlap.

        If we have a set of local sections defining the layout of a set of
        local vectors, and also an `SF` to determine which section points are
        shared and the ownership, we can calculate a global section defining
        the parallel data layout, and the associated global vector.

        This gives negative sizes and offsets to points not owned by this
        process.

        ``includeConstraints`` and ``localOffsets`` parameters of the C API
        are always set to `False`.

        Parameters
        ----------
        sf
            The `SF` describing the parallel layout of the section points
            (leaves are unowned local points).

        See Also
        --------
        petsc.PetscSectionCreateGlobalSection

        """
        cdef Section gsec = Section()
        CHKERR(PetscSectionCreateGlobalSection(self.sec, sf.sf, PETSC_TRUE, PETSC_FALSE, PETSC_FALSE, &gsec.sec))
        return gsec



# --------------------------------------------------------------------

class staticproperty(property):
    def __get__(self, *args, **kwargs):
        return self.fget.__get__(*args, **kwargs)()


cdef object make_enum_class(str class_name, str class_docstring, tuple args):
    cdef dict enum2str = {}
    cdef dict attrs    = {}

    for name, c_enum in args:
        enum2str[c_enum] = name
        attrs[name]      = c_enum

    attrs['__enum2str'] = enum2str
    attrs['__doc__']    = class_docstring
    return type(class_name, (object,), attrs)

DeviceType = make_enum_class(
  "DeviceType",
  """The type of device.

  See Also
  --------
  Device, Device.create, Device.getDeviceType, Device.type, petsc.PetscDeviceType

  """,
  (
    ("HOST"    , PETSC_DEVICE_HOST),
    ("CUDA"    , PETSC_DEVICE_CUDA),
    ("HIP"     , PETSC_DEVICE_HIP),
    ("SYCL"    , PETSC_DEVICE_SYCL),
    ("DEFAULT" , staticproperty(lambda *_, **__: PETSC_DEVICE_DEFAULT()))
  )
)

StreamType = make_enum_class(
  "StreamType",
  """The type of stream.

  See Also
  --------
  DeviceContext, DeviceContext.getStreamType
  DeviceContext.setStreamType, petsc.PetscStreamType

  """,
  (
    ("DEFAULT"                  , PETSC_STREAM_DEFAULT),
    ("NONBLOCKING"              , PETSC_STREAM_NONBLOCKING),
    ("DEFAULT_WITH_BARRIER"     , PETSC_STREAM_DEFAULT_WITH_BARRIER),
    ("NONBLOCKING_WITH_BARRIER" , PETSC_STREAM_NONBLOCKING_WITH_BARRIER),
  )
)

DeviceJoinMode = make_enum_class(
  "DeviceJoinMode",
  """The type of join to perform.

  See Also
  --------
  DeviceContext, DeviceContext.join, DeviceContext.fork
  petsc.PetscDeviceContextJoinMode

  """,
  (
    ("DESTROY" , PETSC_DEVICE_CONTEXT_JOIN_DESTROY),
    ("SYNC"    , PETSC_DEVICE_CONTEXT_JOIN_SYNC),
    ("NO_SYNC" , PETSC_DEVICE_CONTEXT_JOIN_NO_SYNC),
  )
)

# --------------------------------------------------------------------

cdef class Device:
    """The device object.

    Represents a handle to an accelerator (which may be the host).

    See Also
    --------
    DeviceContext, petsc.PetscDevice

    """

    Type = DeviceType

    def __cinit__(self):
        self.device = NULL

    def __dealloc__(self):
        self.destroy()

    @classmethod
    def create(cls, dtype: Type | None = None, device_id: int = DECIDE) -> Device:
        """Create a device object.

        Not collective.

        Parameters
        ----------
        dtype
            The type of device to create (or `None` for the default).

        device_id
            The numeric id of the device to create.

        See Also
        --------
        destroy, petsc.PetscDeviceCreate

        """
        cdef PetscInt        cdevice_id   = asInt(device_id)
        cdef PetscDeviceType cdevice_type = asDeviceType(dtype if dtype is not None else cls.Type.DEFAULT)
        cdef Device          device       = cls()

        CHKERR(PetscDeviceCreate(cdevice_type, cdevice_id, &device.device))
        return device

    def destroy(self) -> None:
        """Destroy a device object.

        Not collective.

        See Also
        --------
        create, petsc.PetscDeviceDestroy

        """
        CHKERR(PetscDeviceDestroy(&self.device))

    def configure(self) -> None:
        """Configure and setup a device object.

        Not collective.

        See Also
        --------
        create, petsc.PetscDeviceConfigure

        """
        CHKERR(PetscDeviceConfigure(self.device))

    def view(self, Viewer viewer=None) -> None:
        """View a device object.

        Collective.

        Parameters
        ----------
        viewer
            A `Viewer` instance or `None` for the default viewer.

        See Also
        --------
        petsc.PetscDeviceView

        """
        cdef PetscViewer vwr = NULL

        if viewer is not None:
            vwr = viewer.vwr
        CHKERR(PetscDeviceView(self.device, vwr))

    def getDeviceType(self) -> str:
        """Return the type of the device.

        Not collective.

        See Also
        --------
        type, petsc.PetscDeviceGetType

        """
        cdef PetscDeviceType cdtype = PETSC_DEVICE_HOST

        CHKERR(PetscDeviceGetType(self.device, &cdtype))
        return toDeviceType(cdtype)

    def getDeviceId(self) -> int:
        """Return the device id.

        Not collective.

        See Also
        --------
        create, petsc.PetscDeviceGetDeviceId

        """
        cdef PetscInt cdevice_id = 0

        CHKERR(PetscDeviceGetDeviceId(self.device, &cdevice_id))
        return toInt(cdevice_id)

    @staticmethod
    def setDefaultType(device_type: Type | str) -> None:
        """Set the device type to be used as the default in subsequent calls to `create`.

        Not collective.

        See Also
        --------
        create, petsc.PetscDeviceSetDefaultDeviceType

        """
        cdef PetscDeviceType cdevice_type = asDeviceType(device_type)

        CHKERR(PetscDeviceSetDefaultDeviceType(cdevice_type))

    property type:
        """The device type."""
        def __get__(self) -> str:
            return self.getDeviceType()

    property device_id:
        """The device id."""
        def __get__(self) -> int:
            return self.getDeviceId()


# --------------------------------------------------------------------

cdef class DeviceContext(Object):
    """DeviceContext object.

    Represents an abstract handle to a device context.

    See Also
    --------
    Device, petsc.PetscDeviceContext

    """
    JoinMode   = DeviceJoinMode
    StreamType = StreamType

    def __cinit__(self):
        self.obj  = <PetscObject*> &self.dctx
        self.dctx = NULL

    def create(self) -> Self:
        """Create an empty DeviceContext.

        Not collective.

        See Also
        --------
        destroy, Device, petsc.PetscDeviceContextCreate

        """
        cdef PetscDeviceContext dctx = NULL
        CHKERR(PetscDeviceContextCreate(&dctx))
        CHKERR(PetscCLEAR(self.obj)); self.dctx = dctx
        return self

    def destroy(self) -> Self:
        """Destroy a device context.

        Not collective.

        See Also
        --------
        create, petsc.PetscDeviceContextDestroy

        """
        CHKERR(PetscDeviceContextDestroy(&self.dctx))
        return self

    def getStreamType(self) -> str:
        """Return the `StreamType`.

        Not collective.

        See Also
        --------
        stream_type, setStreamType, petsc.PetscDeviceContextGetStreamType

        """
        cdef PetscStreamType cstream_type = PETSC_STREAM_DEFAULT

        CHKERR(PetscDeviceContextGetStreamType(self.dctx, &cstream_type))
        return toStreamType(cstream_type)

    def setStreamType(self, stream_type: StreamType | str) -> None:
        """Set the `StreamType`.

        Not collective.

        Parameters
        ----------
        stream_type
            The type of stream to set

        See Also
        --------
        stream_type, getStreamType, petsc.PetscDeviceContextSetStreamType

        """
        cdef PetscStreamType cstream_type = asStreamType(stream_type)

        CHKERR(PetscDeviceContextSetStreamType(self.dctx, cstream_type))

    def getDevice(self) -> Device:
        """Get the `Device` which this instance is attached to.

        Not collective.

        See Also
        --------
        setDevice, device, Device, petsc.PetscDeviceContextGetDevice

        """
        cdef PetscDevice device = NULL

        CHKERR(PetscDeviceContextGetDevice(self.dctx, &device))
        return PyPetscDevice_New(device)

    def setDevice(self, Device device not None) -> None:
        """Set the `Device` which this `DeviceContext` is attached to.

        Collective.

        Parameters
        ----------
        device
            The `Device` to which this instance is attached to.

        See Also
        --------
        getDevice, device, Device, petsc.PetscDeviceContextSetDevice

        """
        cdef PetscDevice cdevice = PyPetscDevice_Get(device)

        CHKERR(PetscDeviceContextSetDevice(self.dctx, cdevice))

    def setUp(self) -> None:
        """Set up the internal data structures for using the device context.

        Not collective.

        See Also
        --------
        create, petsc.PetscDeviceContextSetUp

        """
        CHKERR(PetscDeviceContextSetUp(self.dctx))

    def duplicate(self) -> DeviceContext:
        """Duplicate a the device context.

        Not collective.

        See Also
        --------
        create, petsc.PetscDeviceContextDuplicate

        """
        cdef DeviceContext octx = type(self)()

        CHKERR(PetscDeviceContextDuplicate(self.dctx, &octx.dctx))
        return octx

    def idle(self) -> bool:
        """Return whether the underlying stream for the device context is idle.

        Not collective.

        See Also
        --------
        synchronize, petsc.PetscDeviceContextQueryIdle

        """
        cdef PetscBool is_idle = PETSC_FALSE

        CHKERR(PetscDeviceContextQueryIdle(self.dctx, &is_idle))
        return toBool(is_idle)

    def waitFor(self, other: DeviceContext | None) -> None:
        """Make this instance wait for ``other``.

        Not collective.

        Parameters
        ----------
        other
            The other `DeviceContext` to wait for

        See Also
        --------
        fork, join, petsc.PetscDeviceContextWaitForContext

        """
        cdef PetscDeviceContext cother = NULL

        if other is not None:
            cother = PyPetscDeviceContext_Get(other)
        CHKERR(PetscDeviceContextWaitForContext(self.dctx, cother))

    def fork(self, n: int, stream_type: DeviceContext.StreamType | str | None = None) -> list[DeviceContext]:
        """Create multiple device contexts which are all logically dependent on this one.

        Not collective.

        Parameters
        ----------
        n
            The number of device contexts to create.
        stream_type
            The type of stream of the forked device context.

        Examples
        --------
        The device contexts created must be destroyed using `join`.

        >>> dctx = PETSc.DeviceContext().getCurrent()
        >>> dctxs = dctx.fork(4)
        >>> ... # perform computations
        >>> # we can mix various join modes
        >>> dctx.join(PETSc.DeviceContext.JoinMode.SYNC, dctxs[0:2])
        >>> dctx.join(PETSc.DeviceContext.JoinMode.SYNC, dctxs[2:])
        >>> ... # some more computations and joins
        >>> # dctxs must be all destroyed with joinMode.DESTROY
        >>> dctx.join(PETSc.DeviceContext.JoinMode.DESTROY, dctxs)

        See Also
        --------
        join, waitFor, petsc.PetscDeviceContextFork

        """
        cdef PetscDeviceContext *csubctxs = NULL
        cdef PetscStreamType cstream_type = PETSC_STREAM_DEFAULT
        cdef PetscInt cn = asInt(n)
        cdef list subctxs = []
        if stream_type is None:
            CHKERR(PetscDeviceContextFork(self.dctx, cn, &csubctxs))
        else:
            cstream_type = asStreamType(stream_type)
            CHKERR(PetscDeviceContextForkWithStreamType(self.dctx, cstream_type, cn, &csubctxs))
        # FIXME: without CXX compiler, csubctxs is NULL
        if csubctxs:
            subctxs = [None] * cn
            for i from 0 <= i < cn:
                subctxs[i] = DeviceContext()
                (<DeviceContext?>subctxs[i]).dctx = csubctxs[i]
            CHKERR(PetscFree(csubctxs))
        return subctxs

    def join(self, join_mode: DeviceJoinMode | str, py_sub_ctxs: list[DeviceContext]) -> None:
        """Join a set of device contexts on this one.

        Not collective.

        Parameters
        ----------
        join_mode
            The type of join to perform.
        py_sub_ctxs
            The list of device contexts to join.

        See Also
        --------
        fork, waitFor, petsc.PetscDeviceContextJoin

        """
        cdef PetscDeviceContext *np_subctx = NULL
        cdef PetscDeviceContextJoinMode cjoin_mode = asJoinMode(join_mode)
        cdef Py_ssize_t nctxs = len(py_sub_ctxs)

        CHKERR(PetscMalloc(<size_t>(nctxs) * sizeof(PetscDeviceContext *), &np_subctx))
        for i from 0 <= i < nctxs:
            dctx = py_sub_ctxs[i]
            np_subctx[i] = (<DeviceContext?>dctx).dctx if dctx is not None else NULL
        CHKERR(PetscDeviceContextJoin(self.dctx, <PetscInt>nctxs, cjoin_mode, &np_subctx))

        if cjoin_mode == PETSC_DEVICE_CONTEXT_JOIN_DESTROY:
            # in this case, PETSc destroys the contexts and frees the array
            for i in range(nctxs):
                (<DeviceContext?>py_sub_ctxs[i]).dctx = NULL
        else:
            # we need to free the temporary array
            CHKERR(PetscFree(np_subctx))

    def synchronize(self) -> None:
        """Synchronize a device context.

        Not collective.

        Notes
        -----
        The underlying stream is considered idle after this routine returns,
        i.e. `idle` will return ``True``.

        See Also
        --------
        idle, petsc.PetscDeviceContextSynchronize

        """
        CHKERR(PetscDeviceContextSynchronize(self.dctx))

    def setFromOptions(self, comm: Comm | None = None) -> None:
        """Configure the `DeviceContext` from the options database.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        Sys.getDefaultComm, petsc.PetscDeviceContextSetFromOptions

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)

        CHKERR(PetscDeviceContextSetFromOptions(ccomm, self.dctx))

    @staticmethod
    def getCurrent() -> DeviceContext:
        """Return the current device context.

        Not collective.

        See Also
        --------
        current, setCurrent, petsc.PetscDeviceContextGetCurrentContext

        """
        cdef PetscDeviceContext dctx = NULL

        CHKERR(PetscDeviceContextGetCurrentContext(&dctx))
        return PyPetscDeviceContext_New(dctx)

    @staticmethod
    def setCurrent(dctx: DeviceContext | None) -> None:
        """Set the current device context.

        Not collective.

        Parameters
        ----------
        dctx
            The `DeviceContext` to set as current (or `None` to use
            the default context).

        See Also
        --------
        current, getCurrent, petsc.PetscDeviceContextSetCurrentContext

        """
        cdef PetscDeviceContext cdctx = NULL

        if dctx is not None:
            cdctx = PyPetscDeviceContext_Get(dctx)
        CHKERR(PetscDeviceContextSetCurrentContext(cdctx))

    property stream_type:
        """The stream type."""
        def __get__(self) -> str:
            return self.getStreamType()

        def __set__(self, stype: StreamType | str) -> None:
            self.setStreamType(stype)

    property device:
        """The device associated to the device context."""
        def __get__(self) -> Device:
            return self.getDevice()

        def __set__(self, Device device) -> None:
            self.setDevice(device)

    property current:
        """The current global device context."""
        def __get__(self) -> DeviceContext:
            return self.getCurrent()

        def __set__(self, dctx: DeviceContext | None) -> None:
            self.setCurrent(dctx)

# --------------------------------------------------------------------

del DeviceType
del DeviceJoinMode
del StreamType
del staticproperty



"""
This example demonstrates the use of TAO for Python to solve an
unconstrained minimization problem on a single processor.

We minimize the extended Rosenbrock function::

   sum_{i=0}^{n/2-1} ( alpha*(x_{2i+1}-x_{2i}^2)^2 + (1-x_{2i})^2 )
"""

try: range = xrange
except NameError: pass

# the two lines below are only
# needed to build options database
# from command line arguments
import sys, petsc4py
petsc4py.init(sys.argv)

from petsc4py import PETSc


class AppCtx(object):

    """
    Extended Rosenbrock function.
    """

    def __init__(self, n=2, alpha=99.0):
        self.size  = int(n)
        self.alpha = float(alpha)

    def formObjective(self, tao, x):
        #print 'AppCtx.formObjective()'
        alpha = self.alpha
        nn = self.size // 2
        ff = 0.0
        for i in range(nn):
            t1 = x[2*i+1] - x[2*i] * x[2*i]
            t2 = 1 - x[2*i];
            ff += alpha*t1*t1 + t2*t2;
        return ff

    def formGradient(self, tao, x, G):
        #print 'AppCtx.formGradient()'
        alpha = self.alpha
        nn = self.size // 2
        G.zeroEntries()
        for i in range(nn):
            t1 = x[2*i+1] - x[2*i] * x[2*i]
            t2 = 1 - x[2*i];
            G[2*i]   = -4*alpha*t1*x[2*i] - 2*t2;
            G[2*i+1] = 2*alpha*t1;

    def formObjGrad(self, tao, x, G):
        #print 'AppCtx.formObjGrad()'
        alpha = self.alpha
        nn = self.size // 2
        ff = 0.0
        G.zeroEntries()
        for i in range(nn):
            t1 = x[2*i+1] - x[2*i] * x[2*i]
            t2 = 1 - x[2*i];
            ff += alpha*t1*t1 + t2*t2;
            G[2*i]   = -4*alpha*t1*x[2*i] - 2*t2;
            G[2*i+1] = 2*alpha*t1;
        return ff

    def formHessian(self, tao, x, H, HP):
        #print 'AppCtx.formHessian()'
        alpha = self.alpha
        nn = self.size // 2
        idx = [0, 0]
        v = [[0.0, 0.0],
             [0.0, 0.0]]
        H.zeroEntries()
        for i in range(nn):
            v[1][1] = 2*alpha
            v[0][0] = -4*alpha*(x[2*i+1]-3*x[2*i]*x[2*i]) + 2
            v[1][0] = v[0][1] = -4.0*alpha*x[2*i];
            idx[0] = 2*i
            idx[1] = 2*i+1
            H[idx,idx] = v
        H.assemble()

# access PETSc options database
OptDB = PETSc.Options()


# create user application context
# and configure user parameters
user = AppCtx()
user.size  = OptDB.getInt (    'n', user.size)
user.alpha = OptDB.getReal('alpha', user.alpha)

# create solution vector
x = PETSc.Vec().create(PETSc.COMM_SELF)
x.setSizes(user.size)
x.setFromOptions()

# create Hessian matrix
H = PETSc.Mat().create(PETSc.COMM_SELF)
H.setSizes([user.size, user.size])
H.setFromOptions()
H.setOption(PETSc.Mat.Option.SYMMETRIC, True)
H.setUp()

# pass the following to command line:
#  $ ... -methods nm,lmvm,nls,ntr,cg,blmvm,tron
# to try many methods
methods = OptDB.getString('methods', '')
methods = methods.split(',')
for meth in methods:
    # create TAO Solver
    tao = PETSc.TAO().create(PETSc.COMM_SELF)
    if meth: tao.setType(meth)
    tao.setFromOptions()
    # solve the problem
    tao.setObjectiveGradient(user.formObjGrad)
    tao.setObjective(user.formObjective)
    tao.setGradient(user.formGradient)
    tao.setHessian(user.formHessian, H)
    #app.getKSP().getPC().setFromOptions()
    x.set(0) # zero initial guess
    tao.setSolution(x)
    tao.solve()
    tao.destroy()

## # this is just for testing
## x     = app.getSolution()
## G     = app.getGradient()
## H, HP = app.getHessian()

## f = tao.computeObjective(x)
## tao.computeGradient(x, G)
## f = tao.computeObjectiveGradient(x, G)
## tao.computeHessian(x, H, HP)



# --------------------------------------------------------------------

cdef class Options:
    """The options database object.

    A dictionary-like object to store and operate with
    command line options.

    Parameters
    ----------
    prefix : str, optional
        Optional string to prepend to all the options.

    Examples
    --------
    Create an option database and operate with it.

    >>> from petsc4py import PETSc
    >>> opts = PETSc.Options()
    >>> opts['a'] = 1 # insert the command-line option '-a 1'
    >>> if 'a' in opts: # if the option is present
    >>>     val = opts['a'] # return the option value as 'str'
    >>> a_int = opts.getInt('a') # return the option value as 'int'
    >>> a_bool = opts.getBool('a') # return the option value as 'bool'

    Read command line and use default values.

    >>> from petsc4py import PETSc
    >>> opts = PETSc.Options()
    >>> b_float = opts.getReal('b', 1) # return the value or 1.0 if not present

    Read command line options prepended with a prefix.

    >>> from petsc4py import PETSc
    >>> opts = PETSc.Options('prefix_')
    >>> opts.getString('b', 'some_default_string') # read -prefix_b xxx

    See Also
    --------
    petsc_options

    """

    cdef PetscOptions opt
    cdef object       _prefix

    def __init__(self, prefix = None):
        self.opt = NULL
        self.prefix = prefix

    def __dealloc__(self):
        if self.opt == NULL: return
        CHKERR(PetscOptionsDestroy(&self.opt))

    def __contains__(self, item):
        return self.hasName(item)

    def __getitem__(self, item):
        return self.getString(item)

    def __setitem__(self, item, value):
        self.setValue(item, value)

    def __delitem__(self, item):
        self.delValue(item)

    property prefix:
        """Prefix for options."""
        def __get__(self) -> str:
            return self._prefix

        def __set__(self, prefix):
            self._prefix = getprefix(prefix)

        def __del__(self):
            self._prefix = None
    #

    def create(self) -> Self:
        """Create an options database."""
        if self.opt != NULL: return
        CHKERR(PetscOptionsCreate(&self.opt))
        return self

    def destroy(self) -> Self:
        """Destroy an options database."""
        if self.opt == NULL: return
        CHKERR(PetscOptionsDestroy(&self.opt))
        return self

    def clear(self) -> Self:
        """Clear an options database."""
        if self.opt == NULL: return
        CHKERR(PetscOptionsClear(self.opt))
        return self

    def view(self, Viewer viewer=None) -> None:
        """View the options database.

        Collective.

        Parameters
        ----------
        viewer
            A `Viewer` instance or `None` for the default viewer.

        See Also
        --------
        Viewer, petsc.PetscOptionsView

        """
        cdef PetscViewer vwr = NULL
        if viewer is not None: vwr = viewer.vwr
        CHKERR(PetscOptionsView(self.opt, vwr))

    def prefixPush(self, prefix: str | Options | Object | None) -> None:
        """Push a prefix for the options database.

        Logically collective.

        See Also
        --------
        prefixPop, petsc.PetscOptionsPrefixPush

        """
        prefix = getprefix(prefix)
        cdef const char *cprefix = NULL
        prefix = str2bytes(prefix, &cprefix)
        CHKERR(PetscOptionsPrefixPush(self.opt, cprefix))

    def prefixPop(self) -> None:
        """Pop a prefix for the options database.

        Logically collective.

        See Also
        --------
        prefixPush, petsc.PetscOptionsPrefixPop

        """
        CHKERR(PetscOptionsPrefixPop(self.opt))
    #

    def hasName(self, name: str) -> bool:
        """Return the boolean indicating if the option is in the database."""
        cdef const char *pr = NULL
        cdef const char *nm = NULL
        cdef object unused = getpair(self.prefix, name, &pr, &nm)
        cdef PetscBool flag = PETSC_FALSE
        CHKERR(PetscOptionsHasName(self.opt, pr, nm, &flag))
        return toBool(flag)

    def used(self, name: str) -> bool:
        """Return the boolean indicating if the option was queried from the database."""
        cdef const char *key = NULL
        cdef PetscBool flag = PETSC_FALSE
        name = str2bytes(name, &key)
        CHKERR(PetscOptionsUsed(self.opt, key, &flag))
        return toBool(flag)

    def setValue(self, name: str,
                 value: bool | int | float | Scalar | Sequence[bool] | Sequence[int] | Sequence[float] | Sequence[Scalar] | str) -> None:
        """Set a value for an option.

        Logically collective.

        Parameters
        ----------
        name
            The string identifying the option.
        value
            The option value.

        See Also
        --------
        delValue, petsc.PetscOptionsSetValue

        """
        cdef const char *pr = NULL
        cdef const char *nm = NULL
        cdef object unused = getpair(self.prefix, name, &pr, &nm)
        if pr == NULL:
            option = bytes2str(nm)
        else:
            option = '-%s%s' % (bytes2str(pr), bytes2str(&nm[1]))

        if isinstance(value, ndarray):
            value = value.tolist()
        if isinstance(value, (tuple, list)):
            value = str(value).replace(' ', '').\
                    replace('(', '').replace(')', '').\
                    replace('[', '').replace(']', '')
        elif isinstance(value, bool):
            value = str(value).lower()
        elif value is not None:
            value = str(value)
        cdef const char *key = NULL
        cdef const char *val = NULL
        option = str2bytes(option, &key)
        value  = str2bytes(value,  &val)
        CHKERR(PetscOptionsSetValue(self.opt, key, val))

    def delValue(self, name: str) -> None:
        """Delete an option from the database.

        Logically collective.

        See Also
        --------
        setValue, petsc.PetscOptionsClearValue

        """
        cdef const char *pr = NULL
        cdef const char *nm = NULL
        cdef object unused = getpair(self.prefix, name, &pr, &nm)
        if pr == NULL:
            option = bytes2str(nm)
        else:
            option = '-%s%s' % (bytes2str(pr), bytes2str(&nm[1]))
        cdef const char *key = NULL
        option = str2bytes(option, &key)
        CHKERR(PetscOptionsClearValue(self.opt, key))

    #

    def getBool(self, name: str, default=None) -> bool:
        """Return the boolean value associated with the option.

        Not collective.

        Parameters
        ----------
        name
            The option name.
        default
            The default value.
            If `None`, it raises a `KeyError` if the option is not found.

        See Also
        --------
        getBoolArray, petsc.PetscOptionsGetBool

        """
        return getopt(self.opt, OPT_BOOL, self.prefix, name, default)

    def getBoolArray(self, name: str, default=None) -> ArrayBool:
        """Return the boolean values associated with the option.

        Not collective.

        Parameters
        ----------
        name
            The option name.
        default
            The default value.
            If `None`, it raises a `KeyError` if the option is not found.

        See Also
        --------
        getBool, petsc.PetscOptionsGetBoolArray

        """
        return getopt(self.opt, OPT_BOOLARRAY, self.prefix, name, default)

    def getInt(self, name: str, default=None) -> int:
        """Return the integer value associated with the option.

        Not collective.

        Parameters
        ----------
        name
            The option name.
        default
            The default value.
            If `None`, it raises a `KeyError` if the option is not found.

        See Also
        --------
        getIntArray, petsc.PetscOptionsGetInt

        """
        return getopt(self.opt, OPT_INT, self.prefix, name, default)

    def getIntArray(self, name: str, default=None) -> ArrayInt:
        """Return the integer array associated with the option.

        Not collective.

        Parameters
        ----------
        name
            The option name.
        default
            The default value.
            If `None`, it raises a `KeyError` if the option is not found.

        See Also
        --------
        getInt, petsc.PetscOptionsGetIntArray

        """
        return getopt(self.opt, OPT_INTARRAY, self.prefix, name, default)

    def getReal(self, name: str, default=None) -> float:
        """Return the real value associated with the option.

        Not collective.

        Parameters
        ----------
        name
            The option name.
        default
            The default value.
            If `None`, it raises a `KeyError` if the option is not found.

        See Also
        --------
        getRealArray, petsc.PetscOptionsGetReal

        """
        return getopt(self.opt, OPT_REAL, self.prefix, name, default)

    def getRealArray(self, name: str, default=None) -> ArrayReal:
        """Return the real array associated with the option.

        Not collective.

        Parameters
        ----------
        name
            The option name.
        default
            The default value.
            If `None`, it raises a `KeyError` if the option is not found.

        See Also
        --------
        getReal, petsc.PetscOptionsGetRealArray

        """
        return getopt(self.opt, OPT_REALARRAY, self.prefix, name, default)

    def getScalar(self, name: str, default=None) -> Scalar:
        """Return the scalar value associated with the option.

        Not collective.

        Parameters
        ----------
        name
            The option name.
        default
            The default value.
            If `None`, it raises a `KeyError` if the option is not found.

        See Also
        --------
        getScalarArray, petsc.PetscOptionsGetScalar

        """
        return getopt(self.opt, OPT_SCALAR, self.prefix, name, default)

    def getScalarArray(self, name: str, default=None) -> ArrayScalar:
        """Return the scalar array associated with the option.

        Not collective.

        Parameters
        ----------
        name
            The option name.
        default
            The default value.
            If `None`, it raises a `KeyError` if the option is not found.

        See Also
        --------
        getScalar, petsc.PetscOptionsGetScalarArray

        """
        return getopt(self.opt, OPT_SCALARARRAY, self.prefix, name, default)

    def getString(self, name: str, default=None) -> str:
        """Return the string associated with the option.

        Not collective.

        Parameters
        ----------
        name
            The option name.
        default
            The default value.
            If `None`, it raises a `KeyError` if the option is not found.

        See Also
        --------
        petsc.PetscOptionsGetString

        """
        return getopt(self.opt, OPT_STRING, self.prefix, name, default)

    #

    def insertString(self, string: str) -> None:
        """Insert a string in the options database.

        Logically collective.

        See Also
        --------
        petsc.PetscOptionsInsertString

        """
        cdef const char *cstring = NULL
        string = str2bytes(string, &cstring)
        CHKERR(PetscOptionsInsertString(self.opt, cstring))

    def getAll(self) -> dict[str, str]:
        """Return all the options and their values.

        Not collective.

        See Also
        --------
        petsc.PetscOptionsGetAll

        """
        cdef char *allopts = NULL
        CHKERR(PetscOptionsGetAll(self.opt, &allopts))
        options = bytes2str(allopts)
        CHKERR(PetscFree(allopts))
        return parseopt(options, self.prefix)

# --------------------------------------------------------------------



# This script demonstrates solving a symmetric saddle-point linear system using PETSc and HPDDM preconditioner
# It must be run with exactly 4 MPI processes
# Example run:
#   mpirun -n 4 python3 saddle_point.py -ksp_monitor -ksp_type fgmres -ksp_max_it 10 -ksp_rtol 1e-4 -fieldsplit_ksp_max_it 100 -fieldsplit_pc_hpddm_levels_1_eps_nev 10 -fieldsplit_pc_hpddm_levels_1_st_share_sub_ksp -fieldsplit_pc_hpddm_has_neumann -fieldsplit_pc_hpddm_define_subdomains -fieldsplit_1_pc_hpddm_schur_precondition geneo -fieldsplit_pc_hpddm_coarse_pc_type cholesky -fieldsplit_pc_hpddm_levels_1_sub_pc_type lu -fieldsplit_ksp_type fgmres -fieldsplit_1_pc_hpddm_coarse_correction balanced -fieldsplit_1_pc_hpddm_levels_1_eps_gen_non_hermitian
# For more options, see ${PETSC_DIR}/src/ksp/ksp/tutorials/ex87.c

import sys
import petsc4py
# Initialize PETSc with command-line arguments
petsc4py.init(sys.argv)
from petsc4py import PETSc

# Function to load matrices and index sets from binary files
def mat_and_is_load(prefix, identifier, A, aux_IS, aux_Mat, rank, size):
# Load an index set (IS) from binary file
  sizes = PETSc.IS().load(PETSc.Viewer().createBinary(f"{prefix}{identifier}_sizes_{rank}_{size}.dat", "r", comm = PETSc.COMM_SELF))
# Get indices from the loaded IS
  idx   = sizes.getIndices()
# Set the local and global sizes of the matrix
  A.setSizes([[idx[0], idx[2]], [idx[1], idx[3]]])
# Configure matrix using runtime options
  A.setFromOptions()
# Load matrix A from binary file
  A = A.load(PETSc.Viewer().createBinary(f"{prefix}{identifier}.dat", "r", comm = PETSc.COMM_WORLD))

# Load an index set (IS) from binary file
  aux_IS.load(PETSc.Viewer().createBinary(f"{prefix}{identifier}_is_{rank}_{size}.dat", "r", comm = PETSc.COMM_SELF))
# Load the Neumann matrix of the current process
  aux_Mat.load(PETSc.Viewer().createBinary(f"{prefix}{identifier}_aux_{rank}_{size}.dat", "r", comm = PETSc.COMM_SELF))

# Get the size of the communicator
size = PETSc.COMM_WORLD.getSize()
# Get the rank of the current process
rank = PETSc.COMM_WORLD.getRank()
if size != 4:
  if rank == 0:
    print("This example requires 4 processes")
  quit()

# Problem type (either 'elasticity' or 'stokes')
system_str = PETSc.Options().getString("system", "elasticity")
id_sys = 0 if system_str == "elasticity" else 1
empty_A11 = False
# Lower-left (1,1) block is never zero when problem type is 'elasticity'
if id_sys == 1:
  empty_A11 = PETSc.Options().getBool("empty_A11", False)

# 2-by-2 block structure
A = [None, None, None, None]
# Auxiliary data only for the diagonal blocks
aux_Mat = [None, None]
aux_IS = [None, None]

# Create placeholder objects for the diagonal blocks
A[0] = PETSc.Mat().create(comm = PETSc.COMM_WORLD)
A[0].setFromOptions()
aux_IS[0] = PETSc.IS().create(comm = PETSc.COMM_SELF)
aux_Mat[0] = PETSc.Mat().create(comm = PETSc.COMM_SELF)
A[3] = PETSc.Mat().create(comm = PETSc.COMM_WORLD)
A[3].setFromOptions()
aux_IS[1] = PETSc.IS().create(comm = PETSc.COMM_SELF)
aux_Mat[1] = PETSc.Mat().create(comm = PETSc.COMM_SELF)

# Load directory for input data
load_dir = PETSc.Options().getString("load_dir", "${DATAFILESPATH}/matrices/hpddm/GENEO")
# Specific prefix for each problem
prefix = f"{load_dir}/{ 'B' if id_sys == 1 else 'A' }"

# Diagonal blocks and auxiliary data for PCHPDDM
mat_and_is_load(prefix, "00", A[0], aux_IS[0], aux_Mat[0], rank, size)
mat_and_is_load(prefix, "11", A[3], aux_IS[1], aux_Mat[1], rank, size)

# Coherent off-diagonal (0,1) block
A[2] = PETSc.Mat().create(comm = PETSc.COMM_WORLD)
n, _ = A[0].getLocalSize()
N, _ = A[0].getSize()
m, _ = A[3].getLocalSize()
M, _ = A[3].getSize()
# Set matrix sizes based on the sizes of (0,0) and (1,1) blocks
A[2].setSizes([[m, M], [n, N]])
A[2].setFromOptions()
A[2].load(PETSc.Viewer().createBinary(f"{load_dir}/{ 'B' if id_sys == 1 else 'A' }10.dat", "r", comm = PETSc.COMM_WORLD))
# Create a matrix that behaves likes A[1]' without explicitly assembling it
A[1] = PETSc.Mat().createTranspose(A[2])

# Global MatNest
S = PETSc.Mat().createNest([[A[0], A[1]], [A[2], A[3] if not empty_A11 else None]])

ksp = PETSc.KSP().create(comm = PETSc.COMM_WORLD)
ksp.setOperators(S)
pc = ksp.getPC()

# Use FIELDSPLIT as the outer preconditioner
pc.setType(PETSc.PC.Type.FIELDSPLIT)
# Use SCHUR since there are only two fields
pc.setFieldSplitType(PETSc.PC.CompositeType.SCHUR)
# Use SELF because HPDDM deals with a Schur complement matrix
pc.setFieldSplitSchurPreType(PETSc.PC.FieldSplitSchurPreType.SELF)
# Apply any command-line options (which may override options from the source file)
pc.setFromOptions()
# Setup the outer preconditioner so that one can query the inner preconditioners
pc.setUp()

# Retrieve the inner solvers (associated to the diagonal blocks)
ksp0, ksp1 = pc.getFieldSplitSubKSP()

# Upper-left (0,0) block
pc0 = ksp0.getPC()
# Use HPDDM as the preconditioner
pc0.setType(PETSc.PC.Type.HPDDM)
# Set the index set (local-to-global numbering) and auxiliary matrix (first diagonal block)
pc0.setHPDDMAuxiliaryMat(aux_IS[0], aux_Mat[0])
pc0.setFromOptions()

# Schur complement
pc1 = ksp1.getPC()
# Use HPDDM as the preconditioner
pc1.setType(PETSc.PC.Type.HPDDM)
if not empty_A11:
# Set the index set (local-to-global numbering) and auxiliary matrix (second diagonal block)
# If there is no block (-empty_A11), then these are computed automatically by HPDDM
  pc1.setHPDDMAuxiliaryMat(aux_IS[1], aux_Mat[1])
pc1.setFromOptions()

# Create RHS (b) and solution (x) vectors, load b from file, and solve the system
b, x = S.createVecs()
b.load(PETSc.Viewer().createBinary(f"{load_dir}/rhs_{ 'B' if id_sys == 1 else 'A' }.dat", "r", comm = PETSc.COMM_WORLD))
ksp.setFromOptions()
ksp.solve(b, x)



# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com

# --------------------------------------------------------------------

cdef extern from "<petsc.h>":

    ctypedef struct _p_MPI_Comm
    ctypedef _p_MPI_Comm* MPI_Comm

    ctypedef struct _p_MPI_Op
    ctypedef _p_MPI_Op* MPI_Op

    ctypedef struct _p_MPI_Datatype
    ctypedef _p_MPI_Datatype* MPI_Datatype

    ctypedef struct _p_PetscObject
    ctypedef _p_PetscObject* PetscObject

    struct _p_PetscViewer
    ctypedef _p_PetscViewer* PetscViewer

    struct _p_PetscRandom
    ctypedef _p_PetscRandom* PetscRandom

    struct _n_PetscDevice
    ctypedef _n_PetscDevice* PetscDevice

    struct _p_PetscDeviceContext
    ctypedef _p_PetscDeviceContext* PetscDeviceContext

    struct _p_IS
    ctypedef _p_IS* PetscIS "IS"

    struct _p_ISLocalToGlobalMapping
    ctypedef _p_ISLocalToGlobalMapping* PetscLGMap "ISLocalToGlobalMapping"

    struct _p_PetscSF
    ctypedef _p_PetscSF* PetscSF "PetscSF"

    struct _p_Vec
    ctypedef _p_Vec* PetscVec "Vec"

    struct _p_VecScatter
    ctypedef _p_VecScatter* PetscScatter "VecScatter"

    struct _p_PetscSpace
    ctypedef _p_PetscSpace* PetscSpace "PetscSpace"

    struct _p_PetscDualSpace
    ctypedef _p_PetscDualSpace* PetscDualSpace "PetscDualSpace"

    struct _p_PetscFE
    ctypedef _p_PetscFE* PetscFE "PetscFE"

    struct _p_PetscQuadrature
    ctypedef _p_PetscQuadrature* PetscQuadrature "PetscQuadrature"

    struct _p_DMLabel
    ctypedef _p_DMLabel* PetscDMLabel "DMLabel"

    struct _p_PetscSection
    ctypedef _p_PetscSection* PetscSection

    struct _p_Mat
    ctypedef _p_Mat* PetscMat "Mat"

    struct _p_MatPartitioning
    ctypedef _p_Mat* PetscMatPartitioning "MatPartitioning"

    struct _p_MatNullSpace
    ctypedef _p_MatNullSpace* PetscNullSpace "MatNullSpace"

    struct _p_PC
    ctypedef _p_PC* PetscPC "PC"

    struct _p_KSP
    ctypedef _p_KSP* PetscKSP "KSP"

    struct _p_SNES
    ctypedef _p_SNES* PetscSNES "SNES"

    struct _p_SNESLineSearch
    ctypedef _p_SNESLineSearch* PetscSNESLineSearch "SNESLineSearch"

    struct _p_TS
    ctypedef _p_TS* PetscTS "TS"

    struct _p_TAO "_p_Tao"
    ctypedef _p_TAO* PetscTAO "Tao"

    struct _p_TAOLineSearch "_p_TaoLineSearch"
    ctypedef _p_TAOLineSearch* PetscTAOLineSearch "TaoLineSearch"

    struct _p_AO
    ctypedef _p_AO* PetscAO "AO"

    struct _p_DM
    ctypedef _p_DM* PetscDM "DM"

    struct _p_DMPlexTransform
    ctypedef _p_DMPlexTransform* PetscDMPlexTransform "DMPlexTransform"

    struct _p_DMSwarmCellDM
    ctypedef _p_DMSwarmCellDM* PetscDMSwarmCellDM "DMSwarmCellDM"

    struct _p_PetscDS
    ctypedef _p_PetscDS* PetscDS

    struct _p_PetscPartitioner
    ctypedef _p_PetscPartitioner* PetscPartitioner "PetscPartitioner"

    struct _p_PetscRegressor
    ctypedef _p_PetscRegressor* PetscRegressor "PetscRegressor"

# --------------------------------------------------------------------

ctypedef public api class Comm [
    type   PyPetscComm_Type,
    object PyPetscCommObject,
    ]:
    cdef MPI_Comm comm
    cdef int isdup
    cdef object base

ctypedef public api class Object [
    type   PyPetscObject_Type,
    object PyPetscObjectObject,
    ]:
    cdef __weakref__
    cdef __dummy__
    cdef PetscObject oval
    cdef PetscObject *obj
    cdef object get_attr(self, char name[])
    cdef object set_attr(self, char name[], object attr)
    cdef object get_dict(self)

ctypedef public api class Viewer(Object) [
    type   PyPetscViewer_Type,
    object PyPetscViewerObject,
    ]:
    cdef PetscViewer vwr

ctypedef public api class Random(Object) [
    type   PyPetscRandom_Type,
    object PyPetscRandomObject,
    ]:
    cdef PetscRandom rnd

ctypedef public api class Device [
    type   PyPetscDevice_Type,
    object PyPetscDeviceObject,
    ]:
    cdef PetscDevice device
    cdef object __weakref__

ctypedef public api class DeviceContext(Object) [
    type   PyPetscDeviceContext_Type,
    object PyPetscDeviceContextObject,
    ]:
    cdef PetscDeviceContext dctx

ctypedef public api class IS(Object) [
    type   PyPetscIS_Type,
    object PyPetscISObject,
    ]:
    cdef PetscIS iset

ctypedef public api class LGMap(Object) [
    type   PyPetscLGMap_Type,
    object PyPetscLGMapObject,
    ]:
    cdef PetscLGMap lgm

ctypedef public api class SF(Object) [
    type   PyPetscSF_Type,
    object PyPetscSFObject,
    ]:
    cdef PetscSF sf

ctypedef public api class Vec(Object) [
    type   PyPetscVec_Type,
    object PyPetscVecObject,
    ]:
    cdef PetscVec vec

ctypedef public api class Space(Object) [
    type   PyPetscSpace_Type,
    object PyPetscSpaceObject,
    ]:
    cdef PetscSpace space

ctypedef public api class DualSpace(Object) [
    type   PyPetscDualSpace_Type,
    object PyPetscDualSpaceObject,
    ]:
    cdef PetscDualSpace dualspace

ctypedef public api class FE(Object) [
    type   PyPetscFE_Type,
    object PyPetscFEObject,
    ]:
    cdef PetscFE fe

ctypedef public api class Quad(Object) [
    type   PyPetscQuad_Type,
    object PyPetscQuadObject,
    ]:
    cdef PetscQuadrature quad

ctypedef public api class Scatter(Object) [
    type   PyPetscScatter_Type,
    object PyPetscScatterObject,
    ]:
    cdef PetscScatter sct

ctypedef public api class Section(Object) [
    type   PyPetscSection_Type,
    object PyPetscSectionObject,
    ]:
    cdef PetscSection sec

ctypedef public api class Mat(Object) [
    type   PyPetscMat_Type,
    object PyPetscMatObject,
    ]:
    cdef PetscMat mat

ctypedef public api class MatPartitioning(Object) [
    type   PyPetscMatPartitioning_Type,
    object PyPetscMatPartitioningObject,
    ]:
    cdef PetscMatPartitioning part

ctypedef public api class NullSpace(Object) [
    type   PyPetscNullSpace_Type,
    object PyPetscNullSpaceObject,
    ]:
    cdef PetscNullSpace nsp

ctypedef public api class PC(Object) [
    type   PyPetscPC_Type,
    object PyPetscPCObject,
    ]:
    cdef PetscPC pc

ctypedef public api class KSP(Object) [
    type   PyPetscKSP_Type,
    object PyPetscKSPObject,
    ]:
    cdef PetscKSP ksp

ctypedef public api class SNES(Object) [
    type   PyPetscSNES_Type,
    object PyPetscSNESObject,
    ]:
    cdef PetscSNES snes

ctypedef public api class SNESLineSearch(Object) [
    type   PyPetscSNESLineSearch_Type,
    object PyPetscSNESLineSearchObject,
    ]:
    cdef PetscSNESLineSearch snesls

ctypedef public api class TS(Object) [
    type   PyPetscTS_Type,
    object PyPetscTSObject,
    ]:
    cdef PetscTS ts

ctypedef public api class TAO(Object) [
    type   PyPetscTAO_Type,
    object PyPetscTAOObject,
    ]:
    cdef PetscTAO tao

ctypedef public api class TAOLineSearch(Object) [
    type   PyPetscTAOLineSearch_Type,
    object PyPetscTAOLineSearchObject,
    ]:
    cdef PetscTAOLineSearch taols


ctypedef public api class AO(Object) [
    type   PyPetscAO_Type,
    object PyPetscAOObject,
    ]:
    cdef PetscAO ao

ctypedef public api class DM(Object) [
    type   PyPetscDM_Type,
    object PyPetscDMObject,
    ]:
    cdef PetscDM dm

ctypedef public api class DMPlexTransform(Object) [
    type   PyPetscDMPlexTransform_Type,
    object PyPetscDMPlexTransformObject,
    ]:
    cdef PetscDMPlexTransform tr

ctypedef public api class CellDM(Object) [
    type   PyPetscDMSwarmCellDM_Type,
    object PyPetscDMSwarmCellDMObject,
    ]:
    cdef PetscDMSwarmCellDM cdm

ctypedef public api class DS(Object) [
    type   PyPetscDS_Type,
    object PyPetscDSObject,
    ]:
    cdef PetscDS ds

ctypedef public api class Partitioner(Object) [
    type PyPetscPartitioner_Type,
    object PyPetscPartitionerObject,
    ]:
    cdef PetscPartitioner part

ctypedef public api class DMLabel(Object) [
    type   PyPetscDMLabel_Type,
    object PyPetscDMLabelObject,
    ]:
    cdef PetscDMLabel dmlabel

ctypedef public api class Regressor(Object) [
    type   PyPetscRegressor_Type,
    object PyPetscRegressorObject,
    ]:
    cdef PetscRegressor regressor

# --------------------------------------------------------------------

cdef MPI_Comm GetComm(object, MPI_Comm) except *
cdef MPI_Comm GetCommDefault()

cdef int  PyPetscType_Register(int, type) except -1
cdef type PyPetscType_Lookup(int)

# --------------------------------------------------------------------

cdef extern from * nogil:
    ctypedef enum PetscErrorCode:
        PETSC_SUCCESS
        PETSC_ERR_PLIB
        PETSC_ERR_SUP
        PETSC_ERR_USER
        PETSC_ERR_MEM
        PETSC_ERR_MPI
        PETSC_ERR_PYTHON

    ctypedef enum PetscErrorType:
        PETSC_ERROR_INITIAL
        PETSC_ERROR_REPEAT

cdef PetscErrorCode CHKERR(PetscErrorCode) except PETSC_ERR_PYTHON nogil
cdef PetscErrorCode CHKERRMPI(int) except PETSC_ERR_PYTHON nogil



# Testing TSAdjoint and matrix-free Jacobian
# Basic usage:
#     python vanderpol.py
# Test implicit methods using implicit form:
#     python -implicitform
# Test explicit methods:
#     python -implicitform 0
# Test IMEX methods:
#     python -imexform
# Matrix-free implementations can be enabled with an additional option -mf

import sys
import petsc4py

petsc4py.init(sys.argv)

from petsc4py import PETSc


class VDP:
    n = 2
    comm = PETSc.COMM_SELF

    def __init__(self, mu_=1.0e3, mf_=False, imex_=False):
        self.mu_ = mu_
        self.mf_ = mf_
        self.imex_ = imex_
        if self.mf_:
            self.Jim_ = PETSc.Mat().createDense([self.n, self.n], comm=self.comm)
            self.Jim_.setUp()
            self.JimP_ = PETSc.Mat().createDense([self.n, 1], comm=self.comm)
            self.JimP_.setUp()
            self.Jex_ = PETSc.Mat().createDense([self.n, self.n], comm=self.comm)
            self.Jex_.setUp()
            self.JexP_ = PETSc.Mat().createDense([self.n, 1], comm=self.comm)
            self.JexP_.setUp()

    def initialCondition(self, u):
        mu = self.mu_
        u[0] = 2.0
        u[1] = -2.0 / 3.0 + 10.0 / (81.0 * mu) - 292.0 / (2187.0 * mu * mu)
        u.assemble()

    def evalFunction(self, ts, t, u, f):
        mu = self.mu_
        f[0] = u[1]
        if self.imex_:
            f[1] = 0.0
        else:
            f[1] = mu * ((1.0 - u[0] * u[0]) * u[1] - u[0])
        f.assemble()

    def evalJacobian(self, ts, t, u, A, B):
        if not self.mf_:
            J = A
        else:
            J = self.Jex_
        mu = self.mu_
        J[0, 0] = 0
        J[0, 1] = 1.0
        if self.imex_:
            J[1, 0] = 0
            J[1, 1] = 0
        else:
            J[1, 0] = -mu * (2.0 * u[1] * u[0] + 1.0)
            J[1, 1] = mu * (1.0 - u[0] * u[0])
        J.assemble()
        if A != B:
            B.assemble()

    def evalJacobianP(self, ts, t, u, C):
        if not self.mf_:
            Jp = C
        else:
            Jp = self.JexP_
        if not self.imex_:
            Jp[0, 0] = 0
            Jp[1, 0] = (1.0 - u[0] * u[0]) * u[1] - u[0]
            Jp.assemble()

    def evalIFunction(self, ts, t, u, udot, f):
        mu = self.mu_
        if self.imex_:
            f[0] = udot[0]
        else:
            f[0] = udot[0] - u[1]
        f[1] = udot[1] - mu * ((1.0 - u[0] * u[0]) * u[1] - u[0])
        f.assemble()

    def evalIJacobian(self, ts, t, u, udot, shift, A, B):
        if not self.mf_:
            J = A
        else:
            J = self.Jim_
        mu = self.mu_
        if self.imex_:
            J[0, 0] = shift
            J[0, 1] = 0.0
        else:
            J[0, 0] = shift
            J[0, 1] = -1.0
        J[1, 0] = mu * (2.0 * u[1] * u[0] + 1.0)
        J[1, 1] = shift - mu * (1.0 - u[0] * u[0])
        J.assemble()
        if A != B:
            B.assemble()

    def evalIJacobianP(self, ts, t, u, udot, shift, C):
        if not self.mf_:
            Jp = C
        else:
            Jp = self.JimP_
        Jp[0, 0] = 0
        Jp[1, 0] = u[0] - (1.0 - u[0] * u[0]) * u[1]
        Jp.assemble()


class JacShell:
    def __init__(self, ode):
        self.ode_ = ode

    def mult(self, A, x, y):
        "y <- A * x"
        self.ode_.Jex_.mult(x, y)

    def multTranspose(self, A, x, y):
        "y <- A' * x"
        self.ode_.Jex_.multTranspose(x, y)


class JacPShell:
    def __init__(self, ode):
        self.ode_ = ode

    def multTranspose(self, A, x, y):
        "y <- A' * x"
        self.ode_.JexP_.multTranspose(x, y)


class IJacShell:
    def __init__(self, ode):
        self.ode_ = ode

    def mult(self, A, x, y):
        "y <- A * x"
        self.ode_.Jim_.mult(x, y)

    def multTranspose(self, A, x, y):
        "y <- A' * x"
        self.ode_.Jim_.multTranspose(x, y)


class IJacPShell:
    def __init__(self, ode):
        self.ode_ = ode

    def multTranspose(self, A, x, y):
        "y <- A' * x"
        self.ode_.JimP_.multTranspose(x, y)


OptDB = PETSc.Options()

mu_ = OptDB.getScalar('mu', 1.0e3)
mf_ = OptDB.getBool('mf', False)

implicitform_ = OptDB.getBool('implicitform', False)
imexform_ = OptDB.getBool('imexform', False)

ode = VDP(mu_, mf_, imexform_)

if not mf_:
    Jim = PETSc.Mat().createDense([ode.n, ode.n], comm=ode.comm)
    Jim.setUp()
    JimP = PETSc.Mat().createDense([ode.n, 1], comm=ode.comm)
    JimP.setUp()
    Jex = PETSc.Mat().createDense([ode.n, ode.n], comm=ode.comm)
    Jex.setUp()
    JexP = PETSc.Mat().createDense([ode.n, 1], comm=ode.comm)
    JexP.setUp()
else:
    Jim = PETSc.Mat().create()
    Jim.setSizes([ode.n, ode.n])
    Jim.setType('python')
    shell = IJacShell(ode)
    Jim.setPythonContext(shell)
    Jim.setUp()
    Jim.assemble()
    JimP = PETSc.Mat().create()
    JimP.setSizes([ode.n, 1])
    JimP.setType('python')
    shell = IJacPShell(ode)
    JimP.setPythonContext(shell)
    JimP.setUp()
    JimP.assemble()
    Jex = PETSc.Mat().create()
    Jex.setSizes([ode.n, ode.n])
    Jex.setType('python')
    shell = JacShell(ode)
    Jex.setPythonContext(shell)
    Jex.setUp()
    Jex.assemble()
    JexP = PETSc.Mat().create()
    JexP.setSizes([ode.n, 1])
    JexP.setType('python')
    shell = JacPShell(ode)
    JexP.setPythonContext(shell)
    JexP.setUp()
    JexP.zeroEntries()
    JexP.assemble()

u = PETSc.Vec().createSeq(ode.n, comm=ode.comm)
f = u.duplicate()
adj_u = []
adj_u.append(PETSc.Vec().createSeq(ode.n, comm=ode.comm))
adj_u.append(PETSc.Vec().createSeq(ode.n, comm=ode.comm))
adj_p = []
adj_p.append(PETSc.Vec().createSeq(1, comm=ode.comm))
adj_p.append(PETSc.Vec().createSeq(1, comm=ode.comm))

ts = PETSc.TS().create(comm=ode.comm)
ts.setProblemType(ts.ProblemType.NONLINEAR)

if imexform_:
    ts.setType(ts.Type.ARKIMEX)
    ts.setIFunction(ode.evalIFunction, f)
    ts.setIJacobian(ode.evalIJacobian, Jim)
    ts.setIJacobianP(ode.evalIJacobianP, JimP)
    ts.setRHSFunction(ode.evalFunction, f)
    ts.setRHSJacobian(ode.evalJacobian, Jex)
    ts.setRHSJacobianP(ode.evalJacobianP, JexP)
else:
    if implicitform_:
        ts.setType(ts.Type.CN)
        ts.setIFunction(ode.evalIFunction, f)
        ts.setIJacobian(ode.evalIJacobian, Jim)
        ts.setIJacobianP(ode.evalIJacobianP, JimP)
    else:
        ts.setType(ts.Type.RK)
        ts.setRHSFunction(ode.evalFunction, f)
        ts.setRHSJacobian(ode.evalJacobian, Jex)
        ts.setRHSJacobianP(ode.evalJacobianP, JexP)

ts.setSaveTrajectory()
ts.setTime(0.0)
ts.setTimeStep(0.001)
ts.setMaxTime(0.5)
ts.setMaxSteps(1000)
ts.setExactFinalTime(PETSc.TS.ExactFinalTime.MATCHSTEP)

ts.setFromOptions()
ode.initialCondition(u)
ts.solve(u)

adj_u[0][0] = 1
adj_u[0][1] = 0
adj_u[0].assemble()
adj_u[1][0] = 0
adj_u[1][1] = 1
adj_u[1].assemble()
adj_p[0][0] = 0
adj_p[0].assemble()
adj_p[1][0] = 0
adj_p[1].assemble()

ts.setCostGradients(adj_u, adj_p)

ts.adjointSolve()

adj_u[0].view()
adj_u[1].view()
adj_p[0].view()
adj_p[1].view()


def compute_derp(du, dp):
    print(
        du[1] * (-10.0 / (81.0 * mu_ * mu_) + 2.0 * 292.0 / (2187.0 * mu_ * mu_ * mu_))
        + dp[0]
    )


compute_derp(adj_u[0], adj_p[0])
compute_derp(adj_u[1], adj_p[1])

del ode, Jim, JimP, Jex, JexP, u, f, ts, adj_u, adj_p



import sys, petsc4py
petsc4py.init(sys.argv)

from petsc4py import PETSc
from del2mat import Del2Mat

# this a sequential example
assert PETSc.COMM_WORLD.getSize() == 1

# number of nodes in each direction
# excluding those at the boundary
n = 32
h = 1.0/(n+1) # grid spacing

# setup linear system matrix
A = PETSc.Mat().create()
A.setSizes([n**3, n**3])
A.setType('python')
shell = Del2Mat(n) # shell context
A.setPythonContext(shell)
A.setUp()

# setup linear system vectors
x, b = A.createVecs()
x.set(0.0)
b.set(1.0)

# setup Krylov solver
ksp = PETSc.KSP().create()
pc = ksp.getPC()
ksp.setType('cg')
pc.setType('none')

# iteratively solve linear
# system of equations A*x=b
ksp.setOperators(A)
ksp.setFromOptions()
ksp.solve(b, x)

# scale solution vector to
# account for grid spacing
x.scale(h**2)

OptDB = PETSc.Options()
if OptDB.getBool('plot_mpl', False):
    try:
        from matplotlib import pylab
    except ImportError:
        PETSc.Sys.Print("matplotlib not available")
    else:
        from numpy import mgrid
        X, Y =  mgrid[0:1:1j*n,0:1:1j*n]
        Z = x[...].reshape(n,n,n)[:,:,n/2-2]
        pylab.contourf(X, Y, Z)
        pylab.axis('equal')
        pylab.colorbar()
        pylab.show()



cdef extern from * nogil:

    ctypedef const char* PetscTSType "TSType"
    PetscTSType TSEULER
    PetscTSType TSBEULER
    PetscTSType TSBASICSYMPLECTIC
    PetscTSType TSPSEUDO
    PetscTSType TSCN
    PetscTSType TSSUNDIALS
    PetscTSType TSRK
    PetscTSType TSPYTHON
    PetscTSType TSTHETA
    PetscTSType TSALPHA
    PetscTSType TSALPHA2
    PetscTSType TSGLLE
    PetscTSType TSGLEE
    PetscTSType TSSSP
    PetscTSType TSARKIMEX
    PetscTSType TSDIRK
    PetscTSType TSROSW
    PetscTSType TSEIMEX
    PetscTSType TSMIMEX
    PetscTSType TSBDF
    PetscTSType TSRADAU5
    PetscTSType TSMPRK
    PetscTSType TSDISCGRAD

    ctypedef enum PetscTSProblemType "TSProblemType":
        TS_LINEAR
        TS_NONLINEAR

    ctypedef enum PetscTSEquationType "TSEquationType":
        TS_EQ_UNSPECIFIED
        TS_EQ_EXPLICIT
        TS_EQ_ODE_EXPLICIT
        TS_EQ_DAE_SEMI_EXPLICIT_INDEX1
        TS_EQ_DAE_SEMI_EXPLICIT_INDEX2
        TS_EQ_DAE_SEMI_EXPLICIT_INDEX3
        TS_EQ_DAE_SEMI_EXPLICIT_INDEXHI
        TS_EQ_IMPLICIT
        TS_EQ_ODE_IMPLICIT
        TS_EQ_DAE_IMPLICIT_INDEX1
        TS_EQ_DAE_IMPLICIT_INDEX2
        TS_EQ_DAE_IMPLICIT_INDEX3
        TS_EQ_DAE_IMPLICIT_INDEXHI

    ctypedef enum PetscTSConvergedReason "TSConvergedReason":
        # iterating
        TS_CONVERGED_ITERATING
        # converged
        TS_CONVERGED_TIME
        TS_CONVERGED_ITS
        TS_CONVERGED_USER
        TS_CONVERGED_EVENT
        # diverged
        TS_DIVERGED_NONLINEAR_SOLVE
        TS_DIVERGED_STEP_REJECTED

    ctypedef enum PetscTSExactFinalTimeOption "TSExactFinalTimeOption":
        TS_EXACTFINALTIME_UNSPECIFIED
        TS_EXACTFINALTIME_STEPOVER
        TS_EXACTFINALTIME_INTERPOLATE
        TS_EXACTFINALTIME_MATCHSTEP

    ctypedef PetscErrorCode PetscTSCtxDel(void*)

    ctypedef PetscErrorCode (*PetscTSFunctionFunction)(PetscTS,
                                                       PetscReal,
                                                       PetscVec,
                                                       PetscVec,
                                                       void*) except PETSC_ERR_PYTHON

    ctypedef PetscErrorCode (*PetscTSJacobianFunction)(PetscTS,
                                                       PetscReal,
                                                       PetscVec,
                                                       PetscMat,
                                                       PetscMat,
                                                       void*) except PETSC_ERR_PYTHON

    ctypedef PetscErrorCode (*PetscTSIFunctionFunction)(PetscTS,
                                                        PetscReal,
                                                        PetscVec,
                                                        PetscVec,
                                                        PetscVec,
                                                        void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*PetscTSIJacobianFunction)(PetscTS,
                                                        PetscReal,
                                                        PetscVec,
                                                        PetscVec,
                                                        PetscReal,
                                                        PetscMat,
                                                        PetscMat,
                                                        void*) except PETSC_ERR_PYTHON

    ctypedef PetscErrorCode (*PetscTSIJacobianPFunction)(PetscTS,
                                                         PetscReal,
                                                         PetscVec,
                                                         PetscVec,
                                                         PetscReal,
                                                         PetscMat,
                                                         void*) except PETSC_ERR_PYTHON

    ctypedef PetscErrorCode (*PetscTSI2FunctionFunction)(PetscTS,
                                                         PetscReal,
                                                         PetscVec,
                                                         PetscVec,
                                                         PetscVec,
                                                         PetscVec,
                                                         void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*PetscTSI2JacobianFunction)(PetscTS,
                                                         PetscReal,
                                                         PetscVec,
                                                         PetscVec,
                                                         PetscVec,
                                                         PetscReal,
                                                         PetscReal,
                                                         PetscMat,
                                                         PetscMat,
                                                         void*) except PETSC_ERR_PYTHON

    ctypedef PetscErrorCode (*PetscTSMonitorFunction)(PetscTS,
                                                      PetscInt,
                                                      PetscReal,
                                                      PetscVec,
                                                      void*) except PETSC_ERR_PYTHON

    ctypedef PetscErrorCode (*PetscTSPreStepFunction)  (PetscTS) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*PetscTSPostStepFunction) (PetscTS) except PETSC_ERR_PYTHON

    PetscErrorCode TSCreate(MPI_Comm comm, PetscTS*)
    PetscErrorCode TSClone(PetscTS, PetscTS*)
    PetscErrorCode TSDestroy(PetscTS*)
    PetscErrorCode TSView(PetscTS, PetscViewer)
    PetscErrorCode TSLoad(PetscTS, PetscViewer)

    PetscErrorCode TSSetProblemType(PetscTS, PetscTSProblemType)
    PetscErrorCode TSGetProblemType(PetscTS, PetscTSProblemType*)
    PetscErrorCode TSSetEquationType(PetscTS, PetscTSEquationType)
    PetscErrorCode TSGetEquationType(PetscTS, PetscTSEquationType*)
    PetscErrorCode TSSetType(PetscTS, PetscTSType)
    PetscErrorCode TSGetType(PetscTS, PetscTSType*)

    PetscErrorCode TSSetOptionsPrefix(PetscTS, char[])
    PetscErrorCode TSAppendOptionsPrefix(PetscTS, char[])
    PetscErrorCode TSGetOptionsPrefix(PetscTS, char*[])
    PetscErrorCode TSSetFromOptions(PetscTS)

    PetscErrorCode TSSetSolution(PetscTS, PetscVec)
    PetscErrorCode TSGetSolution(PetscTS, PetscVec*)
    PetscErrorCode TS2SetSolution(PetscTS, PetscVec, PetscVec)
    PetscErrorCode TS2GetSolution(PetscTS, PetscVec*, PetscVec*)

    PetscErrorCode TSGetRHSFunction(PetscTS, PetscVec*, PetscTSFunctionFunction*, void*)
    PetscErrorCode TSGetRHSJacobian(PetscTS, PetscMat*, PetscMat*, PetscTSJacobianFunction*, void**)
    PetscErrorCode TSSetRHSFunction(PetscTS, PetscVec, PetscTSFunctionFunction, void*)
    PetscErrorCode TSSetRHSJacobian(PetscTS, PetscMat, PetscMat, PetscTSJacobianFunction, void*)
    PetscErrorCode TSSetIFunction(PetscTS, PetscVec, PetscTSIFunctionFunction, void*)
    PetscErrorCode TSSetIJacobian(PetscTS, PetscMat, PetscMat, PetscTSIJacobianFunction, void*)
    PetscErrorCode TSSetIJacobianP(PetscTS, PetscMat, PetscTSIJacobianPFunction, void*)
    PetscErrorCode TSGetIFunction(PetscTS, PetscVec*, PetscTSIFunctionFunction*, void*)
    PetscErrorCode TSGetIJacobian(PetscTS, PetscMat*, PetscMat*, PetscTSIJacobianFunction*, void**)
    PetscErrorCode TSSetI2Function(PetscTS, PetscVec, PetscTSI2FunctionFunction, void*)
    PetscErrorCode TSSetI2Jacobian(PetscTS, PetscMat, PetscMat, PetscTSI2JacobianFunction, void*)
    PetscErrorCode TSGetI2Function(PetscTS, PetscVec*, PetscTSI2FunctionFunction*, void**)
    PetscErrorCode TSGetI2Jacobian(PetscTS, PetscMat*, PetscMat*, PetscTSI2JacobianFunction*, void**)
    PetscErrorCode TSRHSSplitSetIS(PetscTS, char[], PetscIS)
    PetscErrorCode TSRHSSplitSetRHSFunction(PetscTS, char[], PetscVec, PetscTSFunctionFunction, void*)
    PetscErrorCode TSRHSSplitSetIFunction(PetscTS, char[], PetscVec, PetscTSIFunctionFunction, void*)
    PetscErrorCode TSRHSSplitSetIJacobian(PetscTS, char[], PetscMat, PetscMat, PetscTSIJacobianFunction, void*)

    PetscErrorCode TSGetKSP(PetscTS, PetscKSP*)
    PetscErrorCode TSGetSNES(PetscTS, PetscSNES*)

    PetscErrorCode TSGetDM(PetscTS, PetscDM*)
    PetscErrorCode TSSetDM(PetscTS, PetscDM)

    PetscErrorCode TSComputeRHSFunction(PetscTS, PetscReal, PetscVec, PetscVec)
    PetscErrorCode TSComputeRHSFunctionLinear(PetscTS, PetscReal, PetscVec, PetscVec, void*)
    PetscErrorCode TSComputeRHSJacobian(PetscTS, PetscReal, PetscVec, PetscMat, PetscMat)
    PetscErrorCode TSComputeRHSJacobianConstant(PetscTS, PetscReal, PetscVec, PetscMat, PetscMat, void*)
    PetscErrorCode TSComputeIFunction(PetscTS, PetscReal, PetscVec, PetscVec, PetscVec, PetscBool)
    PetscErrorCode TSComputeIJacobian(PetscTS, PetscReal, PetscVec, PetscVec, PetscReal, PetscMat, PetscMat, PetscBool)
    PetscErrorCode TSComputeIJacobianP(PetscTS, PetscReal, PetscVec, PetscVec, PetscReal, PetscMat, PetscBool)
    PetscErrorCode TSComputeI2Function(PetscTS, PetscReal, PetscVec, PetscVec, PetscVec, PetscVec)
    PetscErrorCode TSComputeI2Jacobian(PetscTS, PetscReal, PetscVec, PetscVec, PetscVec, PetscReal, PetscReal, PetscMat, PetscMat)

    PetscErrorCode TSSetTime(PetscTS, PetscReal)
    PetscErrorCode TSGetTime(PetscTS, PetscReal*)
    PetscErrorCode TSGetPrevTime(PetscTS, PetscReal*)
    PetscErrorCode TSGetSolveTime(PetscTS, PetscReal*)
    PetscErrorCode TSSetTimeStep(PetscTS, PetscReal)
    PetscErrorCode TSGetTimeStep(PetscTS, PetscReal*)
    PetscErrorCode TSSetStepNumber(PetscTS, PetscInt)
    PetscErrorCode TSGetStepNumber(PetscTS, PetscInt*)
    PetscErrorCode TSSetMaxSteps(PetscTS, PetscInt)
    PetscErrorCode TSGetMaxSteps(PetscTS, PetscInt*)
    PetscErrorCode TSSetMaxTime(PetscTS, PetscReal)
    PetscErrorCode TSGetMaxTime(PetscTS, PetscReal*)
    PetscErrorCode TSSetExactFinalTime(PetscTS, PetscTSExactFinalTimeOption)
    PetscErrorCode TSSetTimeSpan(PetscTS, PetscInt, PetscReal*)
    PetscErrorCode TSSetEvaluationTimes(PetscTS, PetscInt, PetscReal*)
    PetscErrorCode TSGetEvaluationTimes(PetscTS, PetscInt*, const PetscReal**)
    PetscErrorCode TSGetEvaluationSolutions(PetscTS, PetscInt*, const PetscReal**, PetscVec**)

    PetscErrorCode TSSetConvergedReason(PetscTS, PetscTSConvergedReason)
    PetscErrorCode TSGetConvergedReason(PetscTS, PetscTSConvergedReason*)
    PetscErrorCode TSGetSNESIterations(PetscTS, PetscInt*)
    PetscErrorCode TSGetKSPIterations(PetscTS, PetscInt*)
    PetscErrorCode TSGetStepRejections(PetscTS, PetscInt*)
    PetscErrorCode TSSetMaxStepRejections(PetscTS, PetscInt)
    PetscErrorCode TSGetSNESFailures(PetscTS, PetscInt*)
    PetscErrorCode TSSetMaxSNESFailures(PetscTS, PetscInt)
    PetscErrorCode TSSetErrorIfStepFails(PetscTS, PetscBool)
    PetscErrorCode TSSetTolerances(PetscTS, PetscReal, PetscVec, PetscReal, PetscVec)
    PetscErrorCode TSGetTolerances(PetscTS, PetscReal*, PetscVec*, PetscReal*, PetscVec*)

    PetscErrorCode TSMonitorSet(PetscTS, PetscTSMonitorFunction, void*, PetscTSCtxDel*)
    PetscErrorCode TSMonitorCancel(PetscTS)
    PetscErrorCode TSMonitor(PetscTS, PetscInt, PetscReal, PetscVec)

    ctypedef PetscErrorCode (*PetscTSIndicator)(PetscTS, PetscReal, PetscVec, PetscReal[], void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*PetscTSPostEvent)(PetscTS, PetscInt, PetscInt[], PetscReal, PetscVec, PetscBool, void*) except PETSC_ERR_PYTHON

    PetscErrorCode TSSetEventHandler(PetscTS, PetscInt, PetscInt[], PetscBool[], PetscTSIndicator, PetscTSPostEvent, void*)
    PetscErrorCode TSSetEventTolerances(PetscTS, PetscReal, PetscReal[])
    PetscErrorCode TSGetNumEvents(PetscTS, PetscInt*)

    ctypedef PetscErrorCode (*PetscTSAdjointR)(PetscTS, PetscReal, PetscVec, PetscVec, void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*PetscTSAdjointDRDY)(PetscTS, PetscReal, PetscVec, PetscVec[], void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*PetscTSAdjointDRDP)(PetscTS, PetscReal, PetscVec, PetscVec[], void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*PetscTSRHSJacobianP)(PetscTS, PetscReal, PetscVec, PetscMat, void*) except PETSC_ERR_PYTHON

    PetscErrorCode TSSetSaveTrajectory(PetscTS)
    PetscErrorCode TSRemoveTrajectory(PetscTS)
    PetscErrorCode TSSetCostGradients(PetscTS, PetscInt, PetscVec*, PetscVec*)
    PetscErrorCode TSGetCostGradients(PetscTS, PetscInt*, PetscVec**, PetscVec**)
    PetscErrorCode TSCreateQuadratureTS(PetscTS, PetscBool, PetscTS*)
    PetscErrorCode TSGetQuadratureTS(PetscTS, PetscBool*, PetscTS*)
    PetscErrorCode TSGetCostIntegral(PetscTS, PetscVec*)

    PetscErrorCode TSSetRHSJacobianP(PetscTS, PetscMat, PetscTSRHSJacobianP, void*)
    PetscErrorCode TSComputeRHSJacobianP(PetscTS, PetscReal, PetscVec, PetscMat)

    PetscErrorCode TSAdjointSolve(PetscTS)
    PetscErrorCode TSAdjointSetSteps(PetscTS, PetscInt)
    PetscErrorCode TSAdjointStep(PetscTS)
    PetscErrorCode TSAdjointSetUp(PetscTS)
    PetscErrorCode TSAdjointReset(PetscTS)
    PetscErrorCode TSAdjointComputeDRDPFunction(PetscTS, PetscReal, PetscVec, PetscVec*)
    PetscErrorCode TSAdjointComputeDRDYFunction(PetscTS, PetscReal, PetscVec, PetscVec*)
    PetscErrorCode TSAdjointCostIntegral(PetscTS)

    PetscErrorCode TSForwardSetSensitivities(PetscTS, PetscInt, PetscVec*, PetscInt, PetscVec*)
    PetscErrorCode TSForwardGetSensitivities(PetscTS, PetscInt*, PetscVec**, PetscInt*, PetscVec**)
    PetscErrorCode TSForwardSetIntegralGradients(PetscTS, PetscInt, PetscVec *, PetscVec *)
    PetscErrorCode TSForwardGetIntegralGradients(PetscTS, PetscInt*, PetscVec **, PetscVec **)
    PetscErrorCode TSForwardSetRHSJacobianP(PetscTS, PetscVec*, PetscTSCostIntegrandFunction, void*)
    PetscErrorCode TSForwardComputeRHSJacobianP(PetscTS, PetscReal, PetscVec, PetscVec*)
    PetscErrorCode TSForwardSetUp(PetscTS)
    PetscErrorCode TSForwardCostIntegral(PetscTS)
    PetscErrorCode TSForwardStep(PetscTS)

    PetscErrorCode TSSetPreStep(PetscTS, PetscTSPreStepFunction)
    PetscErrorCode TSSetPostStep(PetscTS, PetscTSPostStepFunction)

    PetscErrorCode TSSetUp(PetscTS)
    PetscErrorCode TSReset(PetscTS)
    PetscErrorCode TSStep(PetscTS)
    PetscErrorCode TSRestartStep(PetscTS)
    PetscErrorCode TSRollBack(PetscTS)
    PetscErrorCode TSSolve(PetscTS, PetscVec)
    PetscErrorCode TSInterpolate(PetscTS, PetscReal, PetscVec)
    PetscErrorCode TSPreStage(PetscTS, PetscReal)
    PetscErrorCode TSPostStage(PetscTS, PetscReal, PetscInt, PetscVec*)

    PetscErrorCode TSThetaSetTheta(PetscTS, PetscReal)
    PetscErrorCode TSThetaGetTheta(PetscTS, PetscReal*)
    PetscErrorCode TSThetaSetEndpoint(PetscTS, PetscBool)
    PetscErrorCode TSThetaGetEndpoint(PetscTS, PetscBool*)

    PetscErrorCode TSAlphaSetRadius(PetscTS, PetscReal)
    PetscErrorCode TSAlphaSetParams(PetscTS, PetscReal, PetscReal, PetscReal)
    PetscErrorCode TSAlphaGetParams(PetscTS, PetscReal*, PetscReal*, PetscReal*)

    ctypedef const char* PetscTSRKType "TSRKType"
    PetscTSRKType TSRK1FE
    PetscTSRKType TSRK2A
    PetscTSRKType TSRK2B
    PetscTSRKType TSRK3
    PetscTSRKType TSRK3BS
    PetscTSRKType TSRK4
    PetscTSRKType TSRK5F
    PetscTSRKType TSRK5DP
    PetscTSRKType TSRK5BS
    PetscTSRKType TSRK6VR
    PetscTSRKType TSRK7VR
    PetscTSRKType TSRK8VR

    PetscErrorCode TSRKGetType(PetscTS, PetscTSRKType*)
    PetscErrorCode TSRKSetType(PetscTS, PetscTSRKType)

    ctypedef const char* PetscTSARKIMEXType "TSARKIMEXType"
    PetscTSARKIMEXType TSARKIMEX1BEE
    PetscTSARKIMEXType TSARKIMEXA2
    PetscTSARKIMEXType TSARKIMEXL2
    PetscTSARKIMEXType TSARKIMEXARS122
    PetscTSARKIMEXType TSARKIMEX2C
    PetscTSARKIMEXType TSARKIMEX2D
    PetscTSARKIMEXType TSARKIMEX2E
    PetscTSARKIMEXType TSARKIMEXPRSSP2
    PetscTSARKIMEXType TSARKIMEX3
    PetscTSARKIMEXType TSARKIMEXBPR3
    PetscTSARKIMEXType TSARKIMEXARS443
    PetscTSARKIMEXType TSARKIMEX4
    PetscTSARKIMEXType TSARKIMEX5

    PetscErrorCode TSARKIMEXGetType(PetscTS, PetscTSRKType*)
    PetscErrorCode TSARKIMEXSetType(PetscTS, PetscTSRKType)
    PetscErrorCode TSARKIMEXSetFullyImplicit(PetscTS, PetscBool)
    PetscErrorCode TSARKIMEXSetFastSlowSplit(PetscTS, PetscBool)

    ctypedef const char* PetscTSDIRKType "TSDIRKType"
    PetscTSDIRKType TSDIRKS212
    PetscTSDIRKType TSDIRKES122SAL
    PetscTSDIRKType TSDIRKES213SAL
    PetscTSDIRKType TSDIRKES324SAL
    PetscTSDIRKType TSDIRKES325SAL
    PetscTSDIRKType TSDIRK657A
    PetscTSDIRKType TSDIRKES648SA
    PetscTSDIRKType TSDIRK658A
    PetscTSDIRKType TSDIRKS659A
    PetscTSDIRKType TSDIRK7510SAL
    PetscTSDIRKType TSDIRKES7510SA
    PetscTSDIRKType TSDIRK759A
    PetscTSDIRKType TSDIRKS7511SAL
    PetscTSDIRKType TSDIRK8614A
    PetscTSDIRKType TSDIRK8616SAL
    PetscTSDIRKType TSDIRKES8516SAL

    PetscErrorCode TSDIRKGetType(PetscTS, PetscTSDIRKType*)
    PetscErrorCode TSDIRKSetType(PetscTS, PetscTSDIRKType)

    PetscErrorCode TSPythonSetType(PetscTS, char[])
    PetscErrorCode TSPythonGetType(PetscTS, char*[])

cdef extern from * nogil:
    struct _p_TSAdapt
    ctypedef _p_TSAdapt *PetscTSAdapt "TSAdapt"
    PetscErrorCode TSGetAdapt(PetscTS, PetscTSAdapt*)
    PetscErrorCode TSAdaptGetStepLimits(PetscTSAdapt, PetscReal*, PetscReal*)
    PetscErrorCode TSAdaptSetStepLimits(PetscTSAdapt, PetscReal, PetscReal)
    PetscErrorCode TSAdaptCheckStage(PetscTSAdapt, PetscTS, PetscReal, PetscVec, PetscBool*)

cdef extern from * nogil: # custom.h
    PetscErrorCode TSSetTimeStepNumber(PetscTS, PetscInt)

# -----------------------------------------------------------------------------

cdef inline TS ref_TS(PetscTS ts):
    cdef TS ob = <TS> TS()
    ob.ts = ts
    CHKERR(PetscINCREF(ob.obj))
    return ob

# -----------------------------------------------------------------------------

cdef PetscErrorCode TS_RHSFunction(
    PetscTS   ts,
    PetscReal t,
    PetscVec  x,
    PetscVec  f,
    void      *ctx,
   ) except PETSC_ERR_PYTHON with gil:
    cdef TS  Ts   = ref_TS(ts)
    cdef Vec Xvec = ref_Vec(x)
    cdef Vec Fvec = ref_Vec(f)
    cdef object context = Ts.get_attr('__rhsfunction__')
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (function, args, kargs) = context
    function(Ts, toReal(t), Xvec, Fvec, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode TS_RHSJacobian(
    PetscTS   ts,
    PetscReal t,
    PetscVec  x,
    PetscMat  J,
    PetscMat  P,
    void      *ctx,
   ) except PETSC_ERR_PYTHON with gil:
    cdef TS  Ts   = ref_TS(ts)
    cdef Vec Xvec = ref_Vec(x)
    cdef Mat Jmat = ref_Mat(J)
    cdef Mat Pmat = ref_Mat(P)
    cdef object context = Ts.get_attr('__rhsjacobian__')
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (jacobian, args, kargs) = context
    jacobian(Ts, toReal(t), Xvec, Jmat, Pmat, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode TS_RHSJacobianP(
    PetscTS   ts,
    PetscReal t,
    PetscVec  x,
    PetscMat  J,
    void      *ctx,
   ) except PETSC_ERR_PYTHON with gil:
    cdef TS  Ts   = ref_TS(ts)
    cdef Vec Xvec = ref_Vec(x)
    cdef Mat Jmat = ref_Mat(J)
    cdef object context = Ts.get_attr('__rhsjacobianp__')
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (jacobianp, args, kargs) = context
    jacobianp(Ts, toReal(t), Xvec, Jmat, *args, **kargs)
    return PETSC_SUCCESS

# -----------------------------------------------------------------------------

cdef PetscErrorCode TS_IFunction(
    PetscTS   ts,
    PetscReal t,
    PetscVec  x,
    PetscVec  xdot,
    PetscVec  f,
    void      *ctx,
   ) except PETSC_ERR_PYTHON with gil:
    cdef TS  Ts    = ref_TS(ts)
    cdef Vec Xvec  = ref_Vec(x)
    cdef Vec XDvec = ref_Vec(xdot)
    cdef Vec Fvec  = ref_Vec(f)
    cdef object context = Ts.get_attr('__ifunction__')
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (function, args, kargs) = context
    function(Ts, toReal(t), Xvec, XDvec, Fvec, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode TS_IJacobian(
    PetscTS   ts,
    PetscReal t,
    PetscVec  x,
    PetscVec  xdot,
    PetscReal a,
    PetscMat  J,
    PetscMat  P,
    void      *ctx,
   ) except PETSC_ERR_PYTHON with gil:
    cdef TS   Ts    = ref_TS(ts)
    cdef Vec  Xvec  = ref_Vec(x)
    cdef Vec  XDvec = ref_Vec(xdot)
    cdef Mat  Jmat  = ref_Mat(J)
    cdef Mat  Pmat  = ref_Mat(P)
    cdef object context = Ts.get_attr('__ijacobian__')
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (jacobian, args, kargs) = context
    jacobian(Ts, toReal(t), Xvec, XDvec, toReal(a), Jmat, Pmat, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode TS_IJacobianP(
    PetscTS   ts,
    PetscReal t,
    PetscVec  x,
    PetscVec  xdot,
    PetscReal a,
    PetscMat  J,
    void      *ctx,
   ) except PETSC_ERR_PYTHON with gil:
    cdef TS   Ts    = ref_TS(ts)
    cdef Vec  Xvec  = ref_Vec(x)
    cdef Vec  XDvec = ref_Vec(xdot)
    cdef Mat  Jmat  = ref_Mat(J)
    cdef object context = Ts.get_attr('__ijacobianp__')
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (jacobian, args, kargs) = context
    jacobian(Ts, toReal(t), Xvec, XDvec, toReal(a), Jmat, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode TS_I2Function(
    PetscTS   ts,
    PetscReal t,
    PetscVec  x,
    PetscVec  xdot,
    PetscVec  xdotdot,
    PetscVec  f,
    void      *ctx,
   ) except PETSC_ERR_PYTHON with gil:
    cdef TS  Ts    = ref_TS(ts)
    cdef Vec Xvec  = ref_Vec(x)
    cdef Vec XDvec = ref_Vec(xdot)
    cdef Vec XDDvec = ref_Vec(xdotdot)
    cdef Vec Fvec  = ref_Vec(f)
    cdef object context = Ts.get_attr('__i2function__')
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (function, args, kargs) = context
    function(Ts, toReal(t), Xvec, XDvec, XDDvec, Fvec, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode TS_I2Jacobian(
    PetscTS   ts,
    PetscReal t,
    PetscVec  x,
    PetscVec  xdot,
    PetscVec  xdotdot,
    PetscReal v,
    PetscReal a,
    PetscMat  J,
    PetscMat  P,
    void      *ctx,
   ) except PETSC_ERR_PYTHON with gil:
    cdef TS   Ts    = ref_TS(ts)
    cdef Vec  Xvec  = ref_Vec(x)
    cdef Vec  XDvec = ref_Vec(xdot)
    cdef Vec  XDDvec = ref_Vec(xdotdot)
    cdef Mat  Jmat  = ref_Mat(J)
    cdef Mat  Pmat  = ref_Mat(P)
    cdef object context = Ts.get_attr('__i2jacobian__')
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (jacobian, args, kargs) = context
    jacobian(Ts, toReal(t), Xvec, XDvec, XDDvec, toReal(v), toReal(a), Jmat, Pmat, *args, **kargs)
    return PETSC_SUCCESS

# -----------------------------------------------------------------------------

cdef PetscErrorCode TS_Monitor(
    PetscTS   ts,
    PetscInt  step,
    PetscReal time,
    PetscVec  u,
    void      *ctx,
   ) except PETSC_ERR_PYTHON with gil:
    cdef TS  Ts = ref_TS(ts)
    cdef Vec Vu = ref_Vec(u)
    cdef object monitorlist = Ts.get_attr('__monitor__')
    if monitorlist is None: return PETSC_SUCCESS
    for (monitor, args, kargs) in monitorlist:
        monitor(Ts, toInt(step), toReal(time), Vu, *args, **kargs)
    return PETSC_SUCCESS

# -----------------------------------------------------------------------------

cdef PetscErrorCode TS_Indicator(
    PetscTS   ts,
    PetscReal time,
    PetscVec  u,
    PetscReal fvalue[],
    void      *ctx,
   ) except PETSC_ERR_PYTHON with gil:
    cdef TS  Ts = ref_TS(ts)
    cdef Vec Vu = ref_Vec(u)
    cdef object context = Ts.get_attr('__indicator__')
    if context is None: return PETSC_SUCCESS
    (indicator, args, kargs) = context
    cdef PetscInt nevents = 0
    CHKERR(TSGetNumEvents(ts, &nevents))
    cdef npy_intp s = <npy_intp> nevents
    fvalue_array = PyArray_SimpleNewFromData(1, &s, NPY_PETSC_REAL, fvalue)
    indicator(Ts, toReal(time), Vu, fvalue_array, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode TS_PostEvent(
    PetscTS   ts,
    PetscInt  nevents_zero,
    PetscInt  events_zero[],
    PetscReal time,
    PetscVec  u,
    PetscBool forward,
    void      *ctx,
   ) except PETSC_ERR_PYTHON with gil:
    cdef TS  Ts = ref_TS(ts)
    cdef Vec Vu = ref_Vec(u)
    cdef object context = Ts.get_attr('__postevent__')
    if context is None: return PETSC_SUCCESS
    (postevent, args, kargs) = context
    cdef npy_intp s = <npy_intp> nevents_zero
    events_zero_array = PyArray_SimpleNewFromData(1, &s, NPY_PETSC_INT, events_zero)
    postevent(Ts, events_zero_array, toReal(time), Vu, toBool(forward), *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode TS_PreStep(
    PetscTS ts,
   ) except PETSC_ERR_PYTHON with gil:
    cdef TS Ts = ref_TS(ts)
    (prestep, args, kargs) = Ts.get_attr('__prestep__')
    prestep(Ts, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode TS_PostStep(
    PetscTS ts,
   ) except PETSC_ERR_PYTHON with gil:
    cdef TS Ts = ref_TS(ts)
    (poststep, args, kargs) = Ts.get_attr('__poststep__')
    poststep(Ts, *args, **kargs)
    return PETSC_SUCCESS

# -----------------------------------------------------------------------------



# ------------------------------------------------------------------------
#
#  Poisson problem. This problem is modeled by the partial
#  differential equation
#
#          -Laplacian(u) = 1,  0 < x,y < 1,
#
#  with boundary conditions
#
#           u = 0  for  x = 0, x = 1, y = 0, y = 1
#
#  A finite difference approximation with the usual 5-point stencil
#  is used to discretize the boundary value problem to obtain a
#  nonlinear system of equations. The problem is solved in a 2D
#  rectangular domain, using distributed arrays (DAs) to partition
#  the parallel grid.
#
# ------------------------------------------------------------------------

try: range = xrange
except: pass

import sys, petsc4py
petsc4py.init(sys.argv)

from petsc4py import PETSc

class Poisson2D(object):

    def __init__(self, da):
        assert da.getDim() == 2
        self.da = da
        self.localX  = da.createLocalVec()

    def formRHS(self, B):
        b = self.da.getVecArray(B)
        mx, my = self.da.getSizes()
        hx, hy = [1.0/m for m in [mx, my]]
        (xs, xe), (ys, ye) = self.da.getRanges()
        for j in range(ys, ye):
            for i in range(xs, xe):
                b[i, j] = 1*hx*hy

    def mult(self, mat, X, Y):
        #
        self.da.globalToLocal(X, self.localX)
        x = self.da.getVecArray(self.localX)
        y = self.da.getVecArray(Y)
        #
        mx, my = self.da.getSizes()
        hx, hy = [1.0/m for m in [mx, my]]
        (xs, xe), (ys, ye) = self.da.getRanges()
        for j in range(ys, ye):
            for i in range(xs, xe):
                u = x[i, j] # center
                u_e = u_w = u_n = u_s = 0
                if i > 0:    u_w = x[i-1, j] # west
                if i < mx-1: u_e = x[i+1, j] # east
                if j > 0:    u_s = x[i, j-1] # south
                if j < ny-1: u_n = x[i, j+1] # north
                u_xx = (-u_e + 2*u - u_w)*hy/hx
                u_yy = (-u_n + 2*u - u_s)*hx/hy
                y[i, j] = u_xx + u_yy

OptDB = PETSc.Options()

n  = OptDB.getInt('n', 16)
nx = OptDB.getInt('nx', n)
ny = OptDB.getInt('ny', n)

da = PETSc.DMDA().create([nx, ny], stencil_width=1, setup=False)
da.setFromOptions()
da.setUp()
pde = Poisson2D(da)

x = da.createGlobalVec()
b = da.createGlobalVec()
# A = da.createMat('python')
A = PETSc.Mat().createPython(
    [x.getSizes(), b.getSizes()], comm=da.comm)
A.setPythonContext(pde)
A.setUp()

ksp = PETSc.KSP().create()
ksp.setOperators(A)
ksp.setType('cg')
pc = ksp.getPC()
pc.setType('none')
ksp.setFromOptions()

pde.formRHS(b)
ksp.solve(b, x)

u = da.createNaturalVec()
da.globalToNatural(x, u)

if OptDB.getBool('plot', True):
    draw = PETSc.Viewer.DRAW(x.comm)
    OptDB['draw_pause'] = 1
    draw(x)



# --------------------------------------------------------------------

cdef class Object:
    """Base class wrapping a PETSc object.

    See Also
    --------
    petsc.PetscObject

    """
    # --- special methods ---

    def __cinit__(self):
        self.oval = NULL
        self.obj = &self.oval

    def __dealloc__(self):
        CHKERR(PetscDEALLOC(&self.obj[0]))
        self.obj = NULL

    def __richcmp__(self, other, int op):
        if not isinstance(self,  Object): return NotImplemented
        if not isinstance(other, Object): return NotImplemented
        cdef Object s = self, o = other
        if   op == 2: return (s.obj[0] == o.obj[0])
        elif op == 3: return (s.obj[0] != o.obj[0])
        else: raise TypeError("only '==' and '!='")

    def __bool__(self):
        return self.obj[0] != NULL

    def __copy__(self):
        cdef Object obj = type(self)()
        cdef PetscObject o = self.obj[0]
        if o != NULL:
            CHKERR(PetscObjectReference(o))
        obj.obj[0] = o
        return obj

    def __deepcopy__(self, dict memo):
        cdef object obj_copy = None
        try:
            obj_copy = self.copy
        except AttributeError:
            raise NotImplementedError
        <void>memo # unused
        return obj_copy()

    # --- attribute management ---

    cdef object get_attr(self, char name[]):
        return PetscGetPyObj(self.obj[0], name)

    cdef object set_attr(self, char name[], object attr):
        return PetscSetPyObj(self.obj[0], name, attr)

    cdef object get_dict(self):
        return PetscGetPyDict(self.obj[0], True)

    #

    def view(self, Viewer viewer=None) -> None:
        """Display the object.

        Collective.

        Parameters
        ----------
        viewer
            A `Viewer` instance or `None` for the default viewer.

        See Also
        --------
        petsc.PetscObjectView

        """
        cdef PetscViewer vwr = NULL
        if viewer is not None: vwr = viewer.vwr
        CHKERR(PetscObjectView(self.obj[0], vwr))

    def destroy(self) -> Self:
        """Destroy the object.

        Collective.

        See Also
        --------
        petsc.PetscObjectDestroy

        """
        CHKERR(PetscObjectDestroy(&self.obj[0]))
        return self

    def getType(self) -> str:
        """Return the object type name.

        Not collective.

        See Also
        --------
        petsc.PetscObjectGetType

        """
        cdef const char *cval = NULL
        CHKERR(PetscObjectGetType(self.obj[0], &cval))
        return bytes2str(cval)

    #

    def setOptionsPrefix(self, prefix : str | None) -> None:
        """Set the prefix used for searching for options in the database.

        Logically collective.

        See Also
        --------
        petsc_options, getOptionsPrefix, petsc.PetscObjectSetOptionsPrefix

        """
        cdef const char *cval = NULL
        prefix = str2bytes(prefix, &cval)
        CHKERR(PetscObjectSetOptionsPrefix(self.obj[0], cval))

    def getOptionsPrefix(self) -> str:
        """Return the prefix used for searching for options in the database.

        Not collective.

        See Also
        --------
        petsc_options, setOptionsPrefix, petsc.PetscObjectGetOptionsPrefix

        """
        cdef const char *cval = NULL
        CHKERR(PetscObjectGetOptionsPrefix(self.obj[0], &cval))
        return bytes2str(cval)

    def appendOptionsPrefix(self, prefix: str | None) -> None:
        """Append to the prefix used for searching for options in the database.

        Logically collective.

        See Also
        --------
        petsc_options, setOptionsPrefix, petsc.PetscObjectAppendOptionsPrefix

        """
        cdef const char *cval = NULL
        prefix = str2bytes(prefix, &cval)
        CHKERR(PetscObjectAppendOptionsPrefix(self.obj[0], cval))

    def setFromOptions(self) -> None:
        """Configure the object from the options database.

        Collective.

        Classes that do not implement ``setFromOptions`` use this method
        that, in turn, calls `petsc.PetscObjectSetFromOptions`.

        See Also
        --------
        petsc_options, petsc.PetscObjectSetFromOptions

        """
        CHKERR(PetscObjectSetFromOptions(self.obj[0]))

    def viewFromOptions(self, name : str, Object objpre=None) -> None:
        """View the object via command line options.

        Collective.

        Parameters
        ----------
        name
            The command line option.
        objpre
            Optional object that provides prefix.

        See Also
        --------
        petsc_options, petsc.PetscObjectViewFromOptions

        """
        cdef PetscObject pobj = NULL
        cdef const char *cval = NULL
        pobj = objpre.obj[0] if objpre is not None else NULL
        name = str2bytes(name, &cval)
        CHKERR(PetscObjectViewFromOptions(self.obj[0], pobj, cval))

    def setOptionsHandler(self, handler: PetscOptionsHandlerFunction | None) -> None:
        """Set the callback for processing extra options.

        Logically collective.

        Parameters
        ----------
        handler
            The callback function, called at the end of a ``setFromOptions`` invocation
            for the given class.

        See Also
        --------
        petsc_options, Mat.setFromOptions, KSP.setFromOptions
        petsc.PetscObjectAddOptionsHandler

        """
        if handler is not None:
            CHKERR(PetscObjectAddOptionsHandler(self.obj[0], PetscObjectOptionsHandler_PYTHON, NULL, NULL))
            self.set_attr('__optshandler__', handler)
        else:
            self.set_attr('__optshandler__', None)

    def destroyOptionsHandlers(self) -> None:
        """Clear all the option handlers.

        Collective.

        See Also
        --------
        petsc_options, setOptionsHandler, petsc.PetscObjectDestroyOptionsHandlers

        """
        self.set_attr('__optshandler__', None)
        CHKERR(PetscObjectDestroyOptionsHandlers(self.obj[0]))

    #

    def getComm(self) -> Comm:
        """Return the communicator of the object.

        Not collective.

        See Also
        --------
        petsc.PetscObjectGetComm

        """
        cdef Comm comm = Comm()
        CHKERR(PetscObjectGetComm(self.obj[0], &comm.comm))
        return comm

    def getName(self) -> str:
        """Return the name of the object.

        Not collective.

        See Also
        --------
        petsc.PetscObjectGetName

        """
        cdef const char *cval = NULL
        CHKERR(PetscObjectGetName(self.obj[0], &cval))
        return bytes2str(cval)

    def setName(self, name : str | None) -> None:
        """Associate a name to the object.

        Not collective.

        See Also
        --------
        petsc.PetscObjectSetName

        """
        cdef const char *cval = NULL
        name = str2bytes(name, &cval)
        CHKERR(PetscObjectSetName(self.obj[0], cval))

    def getClassId(self) -> int:
        """Return the class identifier of the object.

        Not collective.

        See Also
        --------
        petsc.PetscObjectGetClassId

        """
        cdef PetscClassId classid = 0
        CHKERR(PetscObjectGetClassId(self.obj[0], &classid))
        return <long>classid

    def getClassName(self) -> str:
        """Return the class name of the object.

        Not collective.

        See Also
        --------
        petsc.PetscObjectGetClassName

        """
        cdef const char *cval = NULL
        CHKERR(PetscObjectGetClassName(self.obj[0], &cval))
        return bytes2str(cval)

    def getRefCount(self) -> int:
        """Return the reference count of the object.

        Not collective.

        See Also
        --------
        petsc.PetscObjectGetReference

        """
        if self.obj[0] == NULL: return 0
        cdef PetscInt refcnt = 0
        CHKERR(PetscObjectGetReference(self.obj[0], &refcnt))
        return toInt(refcnt)

    def getId(self) -> int:
        """Return the unique identifier of the object.

        Not collective.

        See Also
        --------
        petsc.PetscObjectGetId

        """
        cdef PetscObjectId cid = 0
        CHKERR(PetscObjectGetId(self.obj[0], &cid))
        return <long>cid

    # --- general support ---

    def compose(self, name : str | None, Object obj or None) -> None:
        """Associate a PETSc object using a key string.

        Logically collective.

        Parameters
        ----------
        name
            The string identifying the object to be composed.
        obj
            The object to be composed.

        See Also
        --------
        query, petsc.PetscObjectCompose

        """
        cdef const char *cval = NULL
        cdef PetscObject cobj = NULL
        name = str2bytes(name, &cval)
        if obj is not None: cobj = obj.obj[0]
        CHKERR(PetscObjectCompose(self.obj[0], cval, cobj))

    def query(self, name: str) -> Object:
        """Query for the PETSc object associated with a key string.

        Not collective.

        See Also
        --------
        compose, petsc.PetscObjectQuery

        """
        cdef const char *cval = NULL
        cdef PetscObject cobj = NULL
        name = str2bytes(name, &cval)
        CHKERR(PetscObjectQuery(self.obj[0], cval, &cobj))
        if cobj == NULL: return None
        cdef Object obj = subtype_Object(cobj)()
        obj.obj[0] = cobj
        CHKERR(PetscINCREF(obj.obj))
        return obj

    def incRef(self) -> int:
        """Increment the object reference count.

        Logically collective.

        See Also
        --------
        getRefCount, petsc.PetscObjectReference

        """
        cdef PetscObject obj = self.obj[0]
        cdef PetscInt refct = 0
        if obj != NULL:
            CHKERR(PetscObjectReference(obj))
            CHKERR(PetscObjectGetReference(obj, &refct))
        return toInt(refct)

    def decRef(self) -> int:
        """Decrement the object reference count.

        Logically collective.

        See Also
        --------
        getRefCount, petsc.PetscObjectDereference

        """
        cdef PetscObject obj = self.obj[0]
        cdef PetscInt refct = 0
        if obj != NULL:
            CHKERR(PetscObjectGetReference(obj, &refct))
            CHKERR(PetscObjectDereference(obj))
            if refct == 1: self.obj[0] = NULL
            refct -= 1
        return toInt(refct)

    def getAttr(self, name : str) -> object:
        """Return the attribute associated with a given name.

        Not collective.

        See Also
        --------
        setAttr, getDict

        """
        cdef const char *cval = NULL
        name = str2bytes(name, &cval)
        return self.get_attr(<char*>cval)

    def setAttr(self, name : str, attr : object) -> None:
        """Set an the attribute associated with a given name.

        Not collective.

        See Also
        --------
        getAttr, getDict

        """
        cdef const char *cval = NULL
        name = str2bytes(name, &cval)
        self.set_attr(<char*>cval, attr)

    def getDict(self) -> dict:
        """Return the dictionary of attributes.

        Not collective.

        See Also
        --------
        setAttr, getAttr

        """
        return self.get_dict()

    # --- state manipulation ---

    def stateIncrease(self) -> None:
        """Increment the PETSc object state.

        Logically collective.

        See Also
        --------
        stateGet, stateSet, petsc.PetscObjectStateIncrease

        """
        PetscINCSTATE(self.obj)

    def stateGet(self) -> int:
        """Return the PETSc object state.

        Not collective.

        See Also
        --------
        stateSet, stateIncrease, petsc.PetscObjectStateGet

        """
        cdef PetscObjectState state = 0
        CHKERR(PetscObjectStateGet(self.obj[0], &state))
        return <long>state

    def stateSet(self, state : int) -> None:
        """Set the PETSc object state.

        Logically collective.

        See Also
        --------
        stateIncrease, stateGet, petsc.PetscObjectStateSet

        """
        cdef PetscObjectState cstate = asInt(state)
        CHKERR(PetscObjectStateSet(self.obj[0], cstate))

    # --- tab level ---

    def incrementTabLevel(self, tab : int, Object parent=None) -> None:
        """Increment the PETSc object tab level.

        Logically collective.

        See Also
        --------
        setTabLevel, getTabLevel, petsc.PetscObjectIncrementTabLevel

        """
        cdef PetscInt ctab = asInt(tab)
        cdef PetscObject cobj = <PetscObject> NULL if parent is None else parent.obj[0]
        CHKERR(PetscObjectIncrementTabLevel(self.obj[0], cobj, ctab))

    def setTabLevel(self, level : int) -> None:
        """Set the PETSc object tab level.

        Logically collective.

        See Also
        --------
        incrementTabLevel, getTabLevel, petsc.PetscObjectSetTabLevel

        """
        cdef PetscInt clevel = asInt(level)
        CHKERR(PetscObjectSetTabLevel(self.obj[0], clevel))

    def getTabLevel(self) -> None:
        """Return the PETSc object tab level.

        Not collective.

        See Also
        --------
        setTabLevel, incrementTabLevel, petsc.PetscObjectGetTabLevel

        """
        cdef PetscInt clevel = 0
        CHKERR(PetscObjectGetTabLevel(self.obj[0], &clevel))
        return toInt(clevel)

    # --- properties ---

    property type:
        """Object type."""
        def __get__(self) -> str:
            return self.getType()

        def __set__(self, value):
            self.setType(value)

    property prefix:
        """Options prefix."""
        def __get__(self) -> str:
            return self.getOptionsPrefix()

        def __set__(self, value):
            self.setOptionsPrefix(value)

    property comm:
        """The object communicator."""
        def __get__(self) -> Comm:
            return self.getComm()

    property name:
        """The object name."""
        def __get__(self) -> str:
            return self.getName()

        def __set__(self, value):
            self.setName(value)

    property classid:
        """The class identifier."""
        def __get__(self) -> int:
            return self.getClassId()

    property id:
        """The object identifier."""
        def __get__(self) -> int:
            return self.getId()

    property klass:
        """The class name."""
        def __get__(self) -> str:
            return self.getClassName()

    property refcount:
        """Reference count."""
        def __get__(self) -> int:
            return self.getRefCount()

    # --- ctypes support  ---

    property handle:
        """Handle for ctypes support."""
        def __get__(self) -> int:
            cdef PetscObject obj = self.obj[0]
            return PyLong_FromVoidPtr(<void*>obj)

    # --- Fortran support  ---

    property fortran:
        """Fortran handle."""
        def __get__(self) -> int:
            cdef PetscObject obj = self.obj[0]
            return Object_toFortran(obj)

# --------------------------------------------------------------------

include "cyclicgc.pxi"

cdef dict type_registry = {0 : None}
__type_registry__ = type_registry

cdef int PyPetscType_Register(int classid, type cls) except -1:
    global type_registry
    cdef object key = <long>classid
    cdef object value = cls
    cdef const char *dummy = NULL
    if key not in type_registry:
        type_registry[key] = cls
        reg_LogClass(str2bytes(cls.__name__, &dummy),
                     <PetscLogClass>classid)
        TypeEnableGC(<PyTypeObject*>cls)
    else:
        value = type_registry[key]
        if cls is not value:
            raise ValueError(
                "key: %d, cannot register: %s, "
                "already registered: %s" % (key, cls, value))
    return 0

cdef type PyPetscType_Lookup(int classid):
    global type_registry
    cdef object key = <long>classid
    cdef type cls = Object
    try:
        cls = type_registry[key]
    except KeyError:
        cls = Object
    return cls

# --------------------------------------------------------------------



cdef extern from * nogil:

    ctypedef enum PetscDataType:
        PETSC_INT
        PETSC_REAL
        PETSC_SCALAR
        PETSC_COMPLEX
        PETSC_DATATYPE_UNKNOWN

    const char PETSC_AUTHOR_INFO[]
    PetscErrorCode PetscGetVersion(char[], size_t)
    PetscErrorCode PetscGetVersionNumber(PetscInt*, PetscInt*, PetscInt*, PetscInt*)

    PetscErrorCode PetscInitialize(int*, char***, char[], char[])
    PetscErrorCode PetscInitializeNoArguments()
    PetscErrorCode PetscFinalize()
    PetscBool PetscInitializeCalled
    PetscBool PetscFinalizeCalled

    ctypedef PetscErrorCode (*PetscErrorHandlerFunction)(
        MPI_Comm, int, char*, char*, int, PetscErrorType, char*, void*)
    PetscErrorHandlerFunction PetscAttachDebuggerErrorHandler
    PetscErrorHandlerFunction PetscEmacsClientErrorHandler
    PetscErrorHandlerFunction PetscTraceBackErrorHandler
    PetscErrorHandlerFunction PetscMPIAbortErrorHandler
    PetscErrorHandlerFunction PetscAbortErrorHandler
    PetscErrorHandlerFunction PetscIgnoreErrorHandler
    PetscErrorCode PetscPushErrorHandler(PetscErrorHandlerFunction, void*)
    PetscErrorCode PetscPopErrorHandler()
    PetscErrorCode PetscPopSignalHandler()
    PetscErrorCode PetscInfoAllow(PetscBool)
    PetscErrorCode PetscInfoSetFile(char*, char*)

    PetscErrorCode PetscErrorMessage(int, char*[], char**)

    PetscErrorCode PetscSplitOwnership(MPI_Comm, PetscInt*, PetscInt*)
    PetscErrorCode PetscSplitOwnershipBlock(MPI_Comm, PetscInt, PetscInt*, PetscInt*)

    FILE *PETSC_STDOUT
    FILE *PETSC_STDERR

    PetscErrorCode PetscPrintf(MPI_Comm, char[], ...)
    PetscErrorCode PetscVSNPrintf(char*, size_t, const char[], size_t *, va_list)
    PetscErrorCode PetscVFPrintfDefault(FILE*, const char[], va_list)
    PetscErrorCode PetscSynchronizedPrintf(MPI_Comm, char[], ...)
    PetscErrorCode PetscSynchronizedFlush(MPI_Comm, FILE*)

    PetscErrorCode PetscSequentialPhaseBegin(MPI_Comm, int)
    PetscErrorCode PetscSequentialPhaseEnd(MPI_Comm, int)
    PetscErrorCode PetscSleep(PetscReal)

    PetscErrorCode PetscCitationsRegister(const char[], PetscBool*)

    PetscErrorCode PetscHasExternalPackage(const char[], PetscBool*)


cdef inline PetscErrorCode Sys_Sizes(
    object size, object bsize,
    PetscInt *_b,
    PetscInt *_n,
    PetscInt *_N,
   ) except PETSC_ERR_PYTHON:
    # get block size
    cdef PetscInt bs=PETSC_DECIDE, b=PETSC_DECIDE
    if bsize is not None: bs = b = asInt(bsize)
    if bs == PETSC_DECIDE: bs = 1
    # unpack and get local and global sizes
    cdef PetscInt n=PETSC_DECIDE, N=PETSC_DECIDE
    cdef object on, oN
    try:
        on, oN = size
    except (TypeError, ValueError):
        on = None; oN = size
    if on is not None: n = asInt(on)
    if oN is not None: N = asInt(oN)
    # check block, local, and global sizes
    if (bs < 1): raise ValueError(
        "block size %d must be positive" % toInt(bs))
    if n==PETSC_DECIDE and N==PETSC_DECIDE: raise ValueError(
        "local and global sizes cannot be both 'DECIDE'")
    if (n > 0) and (n % bs): raise ValueError(
        "local size %d not divisible by block size %d" %
        (toInt(n), toInt(bs)))
    if (N > 0) and (N % bs): raise ValueError(
        "global size %d not divisible by block size %d" %
        (toInt(N), toInt(bs)))
    # return result to the caller
    if _b != NULL: _b[0] = b
    if _n != NULL: _n[0] = n
    if _N != NULL: _N[0] = N
    return PETSC_SUCCESS

cdef inline PetscErrorCode Sys_Layout(
    MPI_Comm comm,
    PetscInt bs,
    PetscInt *_n,
    PetscInt *_N,
   ) except PETSC_ERR_PYTHON:
    cdef PetscInt n = _n[0]
    cdef PetscInt N = _N[0]
    if bs < 0: bs = 1
    if n  > 0: n = n // bs
    if N  > 0: N = N // bs
    CHKERR(PetscSplitOwnership(comm, &n, &N))
    _n[0] = n * bs
    _N[0] = N * bs
    return PETSC_SUCCESS



# -*- coding: utf-8 -*-
"""
Created on Fri Dec 25 17:03:18 2015

@author: ale
"""

import sys,petsc4py
petsc4py.init(sys.argv)
from petsc4py import PETSc
import numpy as np

dim = 2
if not PETSc.COMM_WORLD.rank:
    coords = np.asarray([[0.0, 0.0],
                         [0.5, 0.0],
                         [1.0, 0.0],
                         [0.0, 0.5],
                         [0.5, 0.5],
                         [1.0, 0.5],
                         [0.0, 1.0],
                         [0.5, 1.0],
                         [1.0, 1.0]], dtype=PETSc.RealType)
    cells = np.asarray([[0,1,4,3],
                        [1,2,5,4],
                        [3,4,7,6],
                        [4,5,8,7]], dtype=PETSc.IntType)
else:
    coords = np.zeros((0, 2), dtype=PETSc.RealType)
    cells = np.zeros((0, 4), dtype=PETSc.IntType)

plex = PETSc.DMPlex().createFromCellList(dim, cells, coords, comm=PETSc.COMM_WORLD)

pStart, pEnd = plex.getChart()
plex.view()
print("pStart, pEnd: ", pStart, pEnd)

# Create section with 1 field with 1 DoF per vertex, edge and cell
numComp = 1
# Start with an empty vector
numDof = [0] * 3
# Field defined on vertices
numDof[0] = 1
# Field defined on edges
numDof[1] = 1
# Field defined on cells
numDof[2] = 1

plex.setNumFields(1)
origSect = plex.createSection(numComp, numDof)
origSect.setFieldName(0, 'TestField')
origSect.setUp()
origSect.view()

plex.setSection(origSect)
origVec = plex.createGlobalVec()
origVec.view()

origVec.setValues(list(range(pStart, pEnd)),list(range(pStart,pEnd)))
origVec.assemblyBegin()
origVec.assemblyEnd()

origVec.view()

if PETSc.COMM_WORLD.size > 1:
    sf = plex.distribute()
    sf.view()

    newSect, newVec = plex.distributeField(sf, origSect, origVec)

else:
    newSect = origSect
    newVec = origVec

newSect.view()
newVec.view()




import sys, petsc4py
petsc4py.init(sys.argv)
from petsc4py import PETSc

# this user class is an application
# context for the nonlinear problem
# at hand; it contains some parameters
# and knows how to compute residuals

class Bratu2D:

    def __init__(self, nx, ny, alpha, impl='python'):
        self.nx = nx # x grid size
        self.ny = ny # y grid size
        self.alpha = alpha
        if impl == 'python':
            from bratu2dnpy import bratu2d
            order = 'c'
        elif impl == 'fortran':
            from bratu2df90 import bratu2d
            order = 'f'
        else:
            raise ValueError('invalid implementation')
        self.compute = bratu2d
        self.order = order

    def evalFunction(self, snes, X, F):
        nx, ny = self.nx, self.ny
        alpha = self.alpha
        order = self.order
        x = X.getArray(readonly=1).reshape(nx, ny, order=order)
        f = F.getArray(readonly=0).reshape(nx, ny, order=order)
        self.compute(alpha, x, f)

# convenience access to
# PETSc options database
OptDB = PETSc.Options()

nx = OptDB.getInt('nx', 32)
ny = OptDB.getInt('ny', nx)
alpha = OptDB.getReal('alpha', 6.8)
impl  = OptDB.getString('impl', 'python')

# create application context
# and PETSc nonlinear solver
appc = Bratu2D(nx, ny, alpha, impl)
snes = PETSc.SNES().create()

# register the function in charge of
# computing the nonlinear residual
f = PETSc.Vec().createSeq(nx*ny)
snes.setFunction(appc.evalFunction, f)

# configure the nonlinear solver
# to use a matrix-free Jacobian
snes.setUseMF(True)
snes.getKSP().setType('cg')
snes.setFromOptions()

# solve the nonlinear problem
b, x = None, f.duplicate()
x.set(0) # zero initial guess
snes.solve(b, x)

if OptDB.getBool('plot', True):
    da = PETSc.DMDA().create([nx,ny])
    u = da.createGlobalVec()
    x.copy(u)
    draw = PETSc.Viewer.DRAW()
    OptDB['draw_pause'] = 1
    draw(u)

if OptDB.getBool('plot_mpl', False):
    try:
        from matplotlib import pylab
    except ImportError:
        PETSc.Sys.Print("matplotlib not available")
    else:
        from numpy import mgrid
        X, Y =  mgrid[0:1:1j*nx,0:1:1j*ny]
        Z = x[...].reshape(nx,ny)
        pylab.figure()
        pylab.contourf(X,Y,Z)
        pylab.colorbar()
        pylab.plot(X.ravel(),Y.ravel(),'.k')
        pylab.axis('equal')
        pylab.show()



# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com


"""
Command line access to the PETSc Options Database.

This module provides command line access to PETSc Options
Database. It outputs a listing of the many PETSc options
indicating option names, default values and descriptions.
Usage::

  $ python -m petsc4py [vec|mat|pc|ksp|snes|ts|tao] [<petsc-option-list>]

"""


def help(args=None):  # noqa: A001
    import shlex
    import sys

    # program name
    try:
        prog = sys.argv[0]
    except Exception:
        prog = getattr(sys, 'executable', 'python')
    # arguments
    if args is None:
        args = sys.argv[1:]
    elif isinstance(args, str):
        args = shlex.split(args)
    else:
        args = [str(a) for a in args]
    if '-help' not in args:
        args.append('-help')
    # import and initialize
    import petsc4py

    petsc4py.init([prog] + args)
    from petsc4py import PETSc

    # help dispatcher
    COMM = PETSc.COMM_SELF
    if 'vec' in args:
        vec = PETSc.Vec().create(comm=COMM)
        vec.setSizes(0)
        vec.setFromOptions()
        vec.destroy()
    if 'mat' in args:
        mat = PETSc.Mat().create(comm=COMM)
        mat.setSizes([0, 0])
        mat.setFromOptions()
        mat.destroy()
    if 'pc' in args:
        pc = PETSc.PC().create(comm=COMM)
        pc.setFromOptions()
        pc.destroy()
    if 'ksp' in args:
        ksp = PETSc.KSP().create(comm=COMM)
        ksp.setFromOptions()
        ksp.destroy()
    if 'snes' in args:
        snes = PETSc.SNES().create(comm=COMM)
        snes.setFromOptions()
        snes.destroy()
    if 'ts' in args:
        ts = PETSc.TS().create(comm=COMM)
        ts.setFromOptions()
        ts.destroy()
    if 'tao' in args:
        tao = PETSc.TAO().create(comm=COMM)
        tao.setFromOptions()
        tao.destroy()
    if 'dmda' in args:
        dmda = PETSc.DMDA().create(comm=COMM)
        dmda.setFromOptions()
        dmda.destroy()
    if 'dmplex' in args:
        dmplex = PETSc.DMPlex().create(comm=COMM)
        dmplex.setFromOptions()
        dmplex.destroy()


if __name__ == '__main__':
    help()




cdef extern from * nogil:

    struct _n_DMInterpolationInfo
    ctypedef _n_DMInterpolationInfo* PetscDMInterpolation "DMInterpolationInfo"

    PetscErrorCode DMInterpolationCreate(MPI_Comm, PetscDMInterpolation*)
    PetscErrorCode DMInterpolationDestroy(PetscDMInterpolation*)
    PetscErrorCode DMInterpolationEvaluate(PetscDMInterpolation, PetscDM, PetscVec, PetscVec)
    PetscErrorCode DMInterpolationGetCoordinates(PetscDMInterpolation, PetscVec*)
    PetscErrorCode DMInterpolationGetDim(PetscDMInterpolation, PetscInt*)
    PetscErrorCode DMInterpolationGetDof(PetscDMInterpolation, PetscInt*)
    PetscErrorCode DMInterpolationGetVector(PetscDMInterpolation, PetscVec*)
    PetscErrorCode DMInterpolationRestoreVector(PetscDMInterpolation, PetscVec*)
    PetscErrorCode DMInterpolationSetDim(PetscDMInterpolation, PetscInt)
    PetscErrorCode DMInterpolationSetDof(PetscDMInterpolation, PetscInt)
    PetscErrorCode DMInterpolationSetUp(PetscDMInterpolation, PetscDM, PetscBool, PetscBool)
